# sage_setup: distribution = sagemath-categories
r"""
Enumerated set of lists of integers with constraints: base classes

- :class:`IntegerListsBackend`: base class for the Cython back-end of
  an enumerated set of lists of integers with specified constraints.

- :class:`Envelope`: a utility class for upper (lower) envelope of a
  function under constraints.
"""

# ****************************************************************************
#       Copyright (C) 2015 Bryan Gillespie <Brg008@gmail.com>
#                          Nicolas M. Thiery <nthiery at users.sf.net>
#                          Anne Schilling <anne@math.ucdavis.edu>
#                          Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


from cpython.object cimport Py_LE, Py_EQ, Py_NE, Py_GE
from sage.misc.constant_function import ConstantFunction
from sage.structure.element cimport RingElement
from sage.rings.integer cimport Integer

Infinity = float('+inf')
MInfinity = float('-inf')


cdef class IntegerListsBackend():
    """
    Base class for the Cython back-end of an enumerated set of lists of
    integers with specified constraints.

    This base implements the basic operations, including checking for
    containment using :meth:`_contains`, but not iteration. For
    iteration, subclass this class and implement an ``_iter()`` method.

    EXAMPLES::

        sage: from sage.combinat.integer_lists.base import IntegerListsBackend
        sage: L = IntegerListsBackend(6, max_slope=-1)
        sage: L._contains([3,2,1])
        True
    """
    def __init__(self,
                 n=None, length=None, *,
                 min_length=0, max_length=Infinity,
                 floor=None, ceiling=None,
                 min_part=0, max_part=Infinity,
                 min_slope=MInfinity, max_slope=Infinity,
                 min_sum=0, max_sum=Infinity):
        """
        Initialize ``self``.

        TESTS::

            sage: from sage.combinat.integer_lists.base import IntegerListsBackend
            sage: C = IntegerListsBackend(2, length=3)
            sage: C = IntegerListsBackend(min_sum=1.4)                                  # needs sage.rings.real_mpfr
            Traceback (most recent call last):
            ...
            TypeError: Attempt to coerce non-integral RealNumber to Integer
            sage: C = IntegerListsBackend(min_sum=Infinity)
            Traceback (most recent call last):
            ...
            TypeError: unable to coerce <class 'sage.rings.infinity.PlusInfinity'> to an integer
        """
        if n is not None:
            min_sum = n
            max_sum = n
        self.min_sum = Integer(min_sum) if min_sum != -Infinity else -Infinity
        self.max_sum = Integer(max_sum) if max_sum != Infinity else Infinity

        if length is not None:
            min_length = length
            max_length = length
        self.min_length = Integer(max(min_length, 0))
        self.max_length = Integer(max_length) if max_length != Infinity else Infinity

        self.min_slope = Integer(min_slope) if min_slope != -Infinity else -Infinity
        self.max_slope = Integer(max_slope) if max_slope !=  Infinity else Infinity

        self.min_part = Integer(min_part) if min_part != -Infinity else -Infinity
        self.max_part = Integer(max_part) if max_part != Infinity else Infinity

        if isinstance(floor, Envelope):
            self.floor = floor
        else:
            if floor is None:
                floor = -Infinity
            elif isinstance(floor, (list, tuple)):
                floor = tuple(Integer(i) for i in floor)
            elif callable(floor):
                pass
            else:
                raise TypeError("floor should be a list, tuple, or function")
            self.floor = Envelope(floor, sign=-1,
                    min_part=self.min_part, max_part=self.max_part,
                    min_slope=self.min_slope, max_slope=self.max_slope,
                    min_length=self.min_length)

        if isinstance(ceiling, Envelope):
            self.ceiling = ceiling
        else:
            if ceiling is None:
                ceiling = Infinity
            elif isinstance(ceiling, (list, tuple)):
                ceiling = tuple(Integer(i) if i != Infinity else Infinity
                                for i in ceiling)
            elif callable(ceiling):
                pass
            else:
                raise ValueError("Unable to parse value of parameter ceiling")
            self.ceiling = Envelope(ceiling, sign=1,
                    min_part=self.min_part, max_part=self.max_part,
                    min_slope=self.min_slope, max_slope=self.max_slope,
                    min_length=self.min_length)

    def __richcmp__(self, other, int op):
        r"""
        Basic comparison function, supporting only checking for
        equality.

        EXAMPLES::

            sage: C = IntegerListsLex(2, length=3).backend
            sage: D = IntegerListsLex(2, length=3).backend; L = list(D._iter())
            sage: E = IntegerListsLex(2, min_length=3).backend
            sage: G = IntegerListsLex(4, length=3).backend
            sage: C >= C
            True
            sage: C == D
            True
            sage: C != D
            False
            sage: C == E
            False
            sage: C != E
            True
            sage: C == None
            False
            sage: C == G
            False
            sage: C <= G
            Traceback (most recent call last):
            ...
            TypeError: IntegerListsBackend can only be compared for equality
        """
        cdef IntegerListsBackend left = <IntegerListsBackend>self
        cdef IntegerListsBackend right = <IntegerListsBackend>other
        equal = (type(left) is type(other) and
            left.min_length == right.min_length and
            left.max_length == right.max_length and
            left.min_sum == right.min_sum and
            left.max_sum == right.max_sum and
            left.min_slope == right.min_slope and
            left.max_slope == right.max_slope and
            left.floor == right.floor and
            left.ceiling == right.ceiling)
        if equal:
            return op == Py_EQ or op == Py_LE or op == Py_GE
        if op == Py_EQ:
            return False
        if op == Py_NE:
            return True
        else:
            raise TypeError("IntegerListsBackend can only be compared for equality")

    def _repr_(self):
        """
        Return the name of this enumerated set.

        EXAMPLES::

            sage: from sage.combinat.integer_lists.base import IntegerListsBackend
            sage: C = IntegerListsBackend(2, length=3)
            sage: C._repr_()
            'Integer lists of sum 2 satisfying certain constraints'
        """
        if self.min_sum == self.max_sum:
            return "Integer lists of sum {} satisfying certain constraints".format(self.min_sum)
        elif self.max_sum == Infinity:
            if self.min_sum == 0:
                return "Integer lists with arbitrary sum satisfying certain constraints"
            else:
                return "Integer lists of sum at least {} satisfying certain constraints".format(self.min_sum)
        else:
            return "Integer lists of sum between {} and {} satisfying certain constraints".format(self.min_sum, self.max_sum)

    def _contains(self, comp):
        """
        Return ``True`` if ``comp`` meets the constraints imposed
        by the arguments.

        EXAMPLES::

            sage: C = IntegerListsLex(n=2, max_length=3, min_slope=0)
            sage: all(l in C for l in C)  # indirect doctest
            True
        """
        if len(comp) < self.min_length or len(comp) > self.max_length:
            return False
        n = sum(comp)
        if n < self.min_sum or n > self.max_sum:
            return False
        for i in range(len(comp)):
            if comp[i] < self.floor(i):
                return False
            if comp[i] > self.ceiling(i):
                return False
        for i in range(len(comp)-1):
            slope = comp[i+1] - comp[i]
            if slope < self.min_slope or slope > self.max_slope:
                return False
        return True

    def __getstate__(self):
        """
        Pickle ``self``.

        EXAMPLES::

            sage: from sage.combinat.integer_lists.base import IntegerListsBackend
            sage: C = IntegerListsBackend(2, length=3)
            sage: C.__getstate__()
            {'ceiling': <sage.combinat.integer_lists.base.Envelope object at ...>,
             'floor': <sage.combinat.integer_lists.base.Envelope object at ...>,
             'max_length': 3,
             'max_part': inf,
             'max_slope': inf,
             'max_sum': 2,
             'min_length': 3,
             'min_part': 0,
             'min_slope': -inf,
             'min_sum': 2}
        """
        return {"min_sum": self.min_sum,
                "max_sum": self.max_sum,
                "min_length": self.min_length,
                "max_length": self.max_length,
                "min_part": self.min_part,
                "max_part": self.max_part,
                "min_slope": self.min_slope,
                "max_slope": self.max_slope,
                "floor": self.floor,
                "ceiling": self.ceiling}

    def __setstate__(self, state):
        """
        Unpickle ``self`` from the state ``state``.

        EXAMPLES::

            sage: from sage.combinat.integer_lists.base import IntegerListsBackend
            sage: C = IntegerListsBackend(2, length=3)
            sage: C == loads(dumps(C))
            True
            sage: C == loads(dumps(C)) # this did fail at some point, really!
            True
            sage: C is loads(dumps(C)) # todo: not implemented
            True
        """
        self.__init__(**state)


cdef class Envelope():
    r"""
    The (currently approximated) upper (lower) envelope of a function
    under the specified constraints.

    INPUT:

    - ``f`` -- a function, list, or tuple; if ``f`` is a list, it is
      considered as the function ``f(i)=f[i]``, completed for larger
      `i` with ``f(i)=max_part``.

    - ``min_part``, ``max_part``, ``min_slope``, ``max_slope``, ...
      as for :class:`IntegerListsLex` (please consult for details).

    - ``sign`` -- (+1 or -1) multiply the input values with ``sign``
      and multiply the output with ``sign``. Setting this to `-1` can
      be used to implement a lower envelope.

    The *upper envelope* `U(f)` of `f` is the (pointwise) largest
    function which is bounded above by `f` and satisfies the
    ``max_part`` and ``max_slope`` conditions. Furthermore, for
    ``i,i+1<min_length``, the upper envelope also satisfies the
    ``min_slope`` condition.

    Upon computing `U(f)(i)`, all the previous values
    for `j\leq i` are computed and cached; in particular `f(i)` will
    be computed at most once for each `i`.

    .. TODO::

        - This class is a good candidate for Cythonization, especially
          to get the critical path in ``__call__`` super fast.

        - To get full envelopes, we would want both the ``min_slope``
          and ``max_slope`` conditions to always be satisfied. This is
          only properly defined for the restriction of `f` to a finite
          interval `0,..,k`, and depends on `k`.

        - This is the core "data structure" of
          ``IntegerListsLex``. Improving the lookahead there
          essentially depends on having functions with a good
          complexity to compute the area below an envelope; and in
          particular how it evolves when increasing the length.

    EXAMPLES::

        sage: from sage.combinat.integer_lists import Envelope

    Trivial upper and lower envelopes::

        sage: f = Envelope([3,2,2])
        sage: [f(i) for i in range(10)]
        [3, 2, 2, inf, inf, inf, inf, inf, inf, inf]
        sage: f = Envelope([3,2,2], sign=-1)
        sage: [f(i) for i in range(10)]
        [3, 2, 2, 0, 0, 0, 0, 0, 0, 0]

    A more interesting lower envelope::

        sage: f = Envelope([4,1,5,3,5], sign=-1, min_part=2, min_slope=-1)
        sage: [f(i) for i in range(10)]
        [4, 3, 5, 4, 5, 4, 3, 2, 2, 2]

    Currently, adding ``max_slope`` has no effect::

        sage: f = Envelope([4,1,5,3,5], sign=-1, min_part=2, min_slope=-1, max_slope=0)
        sage: [f(i) for i in range(10)]
        [4, 3, 5, 4, 5, 4, 3, 2, 2, 2]

    unless ``min_length`` is large enough::

        sage: f = Envelope([4,1,5,3,5], sign=-1, min_part=2, min_slope=-1, max_slope=0, min_length=2)
        sage: [f(i) for i in range(10)]
        [4, 3, 5, 4, 5, 4, 3, 2, 2, 2]

        sage: f = Envelope([4,1,5,3,5], sign=-1, min_part=2, min_slope=-1, max_slope=0, min_length=4)
        sage: [f(i) for i in range(10)]
        [5, 5, 5, 4, 5, 4, 3, 2, 2, 2]

        sage: f = Envelope([4,1,5,3,5], sign=-1, min_part=2, min_slope=-1, max_slope=0, min_length=5)
        sage: [f(i) for i in range(10)]
        [5, 5, 5, 5, 5, 4, 3, 2, 2, 2]

    A non trivial upper envelope::

        sage: f = Envelope([9,1,5,4], max_part=7, max_slope=2)
        sage: [f(i) for i in range(10)]
        [7, 1, 3, 4, 6, 7, 7, 7, 7, 7]

    TESTS::

        sage: f = Envelope(3, min_slope=1)
        sage: [f(i) for i in range(10)]
        [3, 3, 3, 3, 3, 3, 3, 3, 3, 3]

        sage: f = Envelope(3, min_slope=1, min_length=5)
        sage: [f(i) for i in range(10)]
        [-1, 0, 1, 2, 3, 3, 3, 3, 3, 3]

        sage: f = Envelope(3, sign=-1, min_slope=1)
        sage: [f(i) for i in range(10)]
        [3, 4, 5, 6, 7, 8, 9, 10, 11, 12]

        sage: f = Envelope(3, sign=-1, max_slope=-1, min_length=4)
        sage: [f(i) for i in range(10)]
        [6, 5, 4, 3, 3, 3, 3, 3, 3, 3]
    """
    def __init__(self, f, *,
                 min_part=0, max_part=Infinity,
                 min_slope=MInfinity, max_slope=Infinity,
                 min_length=0, max_length=Infinity, sign=1):
        r"""
        Initialize this envelope.

        TESTS::

            sage: from sage.combinat.integer_lists import Envelope
            sage: f = Envelope(3, sign=-1, max_slope=-1, min_length=4)
            sage: f.sign
            -1
            sage: f.max_part
            -3
            sage: f.max_slope
            inf
            sage: f.min_slope
            1
            sage: TestSuite(f).run(skip='_test_pickling')
            sage: Envelope(3, sign=1/3, max_slope=-1, min_length=4)
            Traceback (most recent call last):
            ...
            TypeError: no conversion of this rational to integer
            sage: Envelope(3, sign=-2, max_slope=-1, min_length=4)
            Traceback (most recent call last):
            ...
            ValueError: sign should be +1 or -1
        """
        # self.sign = sign for the output values (the sign change for
        # f is handled here in __init__)
        self.sign = Integer(sign)
        if self.sign == 1:
            self.max_part = max_part
            self.min_slope = min_slope
            self.max_slope = max_slope
            if max_part == 0:
                # This uses that all entries are nonnegative.
                # This is not for speed optimization but for
                # setting the limit start and avoid hangs.
                # See #17979: comment 389
                f = Integer(0)
        elif self.sign == -1:
            self.max_part = -min_part
            self.min_slope = -max_slope
            self.max_slope = -min_slope
        else:
            raise ValueError("sign should be +1 or -1")

        # Handle different types of f and multiply f with sign
        if isinstance(f, RingElement) or f == Infinity or f == -Infinity:
            limit_start = 0
            self.max_part = min(self.sign * f, self.max_part)
            f = ConstantFunction(Infinity)
        elif isinstance(f, (list, tuple)):
            limit_start = len(f)
            f_tab = [self.sign * i for i in f]
            f = lambda k: f_tab[k] if k < len(f_tab) else Infinity
        else:
            g = f
            f = lambda k: self.sign * g(k)
            # At this point, this is not really used
            limit_start = Infinity

        self.f = f
        # For i >= limit_start, f is constant
        # This does not necessarily means that self is constant!
        self.f_limit_start = limit_start
        self.precomputed = []

        if min_length > 0:
            self(min_length-1)
            for i in range(min_length-1,0,-1):
                self.precomputed[i-1] = min(self.precomputed[i-1], self.precomputed[i] - self.min_slope)

    def __richcmp__(self, other, int op):
        r"""
        Basic comparison function, supporting only checking for
        equality.

        EXAMPLES::

            sage: from sage.combinat.integer_lists import Envelope
            sage: f = Envelope([3,2,2])
            sage: g = Envelope([3,2,2])
            sage: h = Envelope([3,2,2], min_part=2)
            sage: f == f, f == h, f == None
            (True, False, False)
            sage: f < f, f != h, f != None
            (False, True, True)

        This would be desirable::

            sage: f == g          # todo: not implemented
            True
        """
        cdef Envelope left = <Envelope>self
        cdef Envelope right = <Envelope>other
        equal = (type(left) is type(other) and
            left.sign == right.sign and
            left.f == right.f and
            left.f_limit_start == right.f_limit_start and
            left.max_part == right.max_part and
            left.min_slope == right.min_slope and
            left.max_slope == right.max_slope)
        if equal:
            return op == Py_EQ or op == Py_LE or op == Py_GE
        if op == Py_EQ:
            return False
        if op == Py_NE:
            return True
        else:
            raise TypeError("Envelopes can only be compared for equality")

    def limit_start(self):
        """
        Return from which `i` on the bound returned by ``limit`` holds.

        .. SEEALSO:: :meth:`limit` for the precise specifications.

        EXAMPLES::

            sage: from sage.combinat.integer_lists import Envelope
            sage: Envelope([4,1,5]).limit_start()
            3
            sage: Envelope([4,1,5], sign=-1).limit_start()
            3

            sage: Envelope([4,1,5], max_part=2).limit_start()
            3

            sage: Envelope(4).limit_start()
            0
            sage: Envelope(4, sign=-1).limit_start()
            0

            sage: Envelope(lambda x: 3).limit_start() == Infinity
            True
            sage: Envelope(lambda x: 3, max_part=2).limit_start() == Infinity
            True

            sage: Envelope(lambda x: 3, sign=-1, min_part=2).limit_start() == Infinity
            True
        """
        return self.f_limit_start

    def limit(self):
        r"""
        Return a bound on the limit of ``self``.

        OUTPUT: nonnegative integer or `\infty`

        This returns some upper bound for the accumulation points of
        this upper envelope. For a lower envelope, a lower bound is
        returned instead.

        In particular this gives a bound for the value of ``self`` at
        `i` for `i` large enough. Special case: for a lower envelop,
        and when the limit is `\infty`, the envelope is guaranteed to
        tend to `\infty` instead.

        When ``s=self.limit_start()`` is finite, this bound is
        guaranteed to be valid for `i>=s`.

        Sometimes it's better to have a loose bound that starts early;
        sometimes the converse holds. At this point which specific
        bound and starting point is returned is not set in stone, in
        order to leave room for later optimizations.

        EXAMPLES::

            sage: from sage.combinat.integer_lists import Envelope
            sage: Envelope([4,1,5]).limit()
            inf
            sage: Envelope([4,1,5], max_part=2).limit()
            2
            sage: Envelope([4,1,5], max_slope=0).limit()
            1
            sage: Envelope(lambda x: 3, max_part=2).limit()
            2

        Lower envelopes::

            sage: Envelope(lambda x: 3, min_part=2, sign=-1).limit()
            2
            sage: Envelope([4,1,5], min_slope=0, sign=-1).limit()
            5
            sage: Envelope([4,1,5], sign=-1).limit()
            0

        .. SEEALSO:: :meth:`limit_start`
        """
        if self.limit_start() < Infinity and self.max_slope <= 0:
            return self(self.limit_start())
        else:
            return self.max_part * self.sign

    def __call__(self, Py_ssize_t k):
        """
        Return the value of this envelope at `k`.

        EXAMPLES::

            sage: from sage.combinat.integer_lists import Envelope
            sage: f = Envelope([4,1,5,3,5])
            sage: f.__call__(2)
            5
            sage: [f(i) for i in range(10)]
            [4, 1, 5, 3, 5, inf, inf, inf, inf, inf]

        .. NOTE::

            See the documentation of :class:`Envelope` for tests and
            examples.
        """
        if k >= len(self.precomputed):
            for i in range(len(self.precomputed), k+1):
                value = min(self.f(i), self.max_part)
                if i > 0:
                    value = min(value, self.precomputed[i-1] + self.max_slope)
                self.precomputed.append(value)
        return self.precomputed[k] * self.sign

    def adapt(self, m, j):
        """
        Return this envelope adapted to an additional local constraint.

        INPUT:

        - ``m`` -- nonnegative integer (starting value)

        - ``j`` -- nonnegative integer (position)

        This method adapts this envelope to the additional local
        constraint imposed by having a part `m` at position `j`.
        Namely, this returns a function which computes, for any `i>j`,
        the minimum of the ceiling function and the value restriction
        given by the slope conditions.

        EXAMPLES::

            sage: from sage.combinat.integer_lists import Envelope
            sage: f = Envelope(3)
            sage: g = f.adapt(1,1)
            sage: g is f
            True
            sage: [g(i) for i in range(10)]
            [3, 3, 3, 3, 3, 3, 3, 3, 3, 3]

            sage: f = Envelope(3, max_slope=1)
            sage: g = f.adapt(1,1)
            sage: [g(i) for i in range(10)]
            [0, 1, 2, 3, 3, 3, 3, 3, 3, 3]

        Note that, in both cases above, the adapted envelope is only
        guaranteed to be valid for `i>j`! This is to leave potential
        room in the future for sharing similar adapted envelopes::

            sage: g = f.adapt(0,0)
            sage: [g(i) for i in range(10)]
            [0, 1, 2, 3, 3, 3, 3, 3, 3, 3]

            sage: g = f.adapt(2,2)
            sage: [g(i) for i in range(10)]
            [0, 1, 2, 3, 3, 3, 3, 3, 3, 3]

            sage: g = f.adapt(3,3)
            sage: [g(i) for i in range(10)]
            [0, 1, 2, 3, 3, 3, 3, 3, 3, 3]

        Now with a lower envelope::

            sage: f = Envelope(1, sign=-1, min_slope=-1)
            sage: g = f.adapt(2,2)
            sage: [g(i) for i in range(10)]
            [4, 3, 2, 1, 1, 1, 1, 1, 1, 1]
            sage: g = f.adapt(1,3)
            sage: [g(i) for i in range(10)]
            [4, 3, 2, 1, 1, 1, 1, 1, 1, 1]
        """
        if self.max_slope == Infinity:
            return self
        m *= self.sign
        m = m - j * self.max_slope
        return lambda i: self.sign * min(m + i*self.max_slope, self.sign*self(i) )

    def __reduce__(self):
        """
        Pickle ``self``.

        EXAMPLES::

            sage: from sage.combinat.integer_lists import Envelope
            sage: h = Envelope(3, min_part=2)
            sage: loads(dumps(h)) == h
            True
        """
        args = (type(self),
                self.sign, self.f, self.f_limit_start, self.precomputed,
                self.max_part, self.min_slope, self.max_slope)
        return _unpickle_Envelope, args


def _unpickle_Envelope(type t, _sign, _f, _f_limit_start, _precomputed,
        _max_part, _min_slope, _max_slope):
    """
    Internal function to support pickling for :class:`Envelope`.

    EXAMPLES::

        sage: from sage.combinat.integer_lists.base import Envelope, _unpickle_Envelope
        sage: _unpickle_Envelope(Envelope,
        ....:     1, lambda i:i, Infinity, [], 4, -1, 3)
        <sage.combinat.integer_lists.base.Envelope object at ...>
    """
    cdef Envelope self = t.__new__(t)
    self.sign = _sign
    self.f = _f
    self.f_limit_start = _f_limit_start
    self.precomputed = _precomputed
    self.max_part = _max_part
    self.min_slope = _min_slope
    self.max_slope = _max_slope
    return self



# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.modules
"""
Ideals of non-commutative rings

Generic implementation of one- and two-sided ideals of non-commutative rings.

AUTHOR:

- Simon King (2011-03-21), <simon.king@uni-jena.de>, :issue:`7797`.

EXAMPLES::

    sage: MS = MatrixSpace(ZZ,2,2)
    sage: MS*MS([0,1,-2,3])
    Left Ideal
    (
      [ 0  1]
      [-2  3]
    )
     of Full MatrixSpace of 2 by 2 dense matrices over Integer Ring
    sage: MS([0,1,-2,3])*MS
    Right Ideal
    (
      [ 0  1]
      [-2  3]
    )
     of Full MatrixSpace of 2 by 2 dense matrices over Integer Ring
    sage: MS*MS([0,1,-2,3])*MS
    Twosided Ideal
    (
      [ 0  1]
      [-2  3]
    )
     of Full MatrixSpace of 2 by 2 dense matrices over Integer Ring

See :mod:`~sage.algebras.letterplace.letterplace_ideal` for a more
elaborate implementation in the special case of ideals in free
algebras.

TESTS::

    sage: # needs sage.combinat
    sage: A = SteenrodAlgebra(2)
    sage: IL = A*[A.1+A.2,A.1^2]; IL
    Left Ideal (Sq(2) + Sq(4), Sq(1,1)) of mod 2 Steenrod algebra, milnor basis
    sage: TestSuite(IL).run(skip=['_test_category'],verbose=True)
    running ._test_eq() . . . pass
    running ._test_new() . . . pass
    running ._test_not_implemented_methods() . . . pass
    running ._test_pickling() . . . pass
"""
# ****************************************************************************
#       Copyright (C) 2011 Simon King <simon.king@uni-jena.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.element cimport MonoidElement
from sage.structure.parent cimport Parent
from sage.categories.monoids import Monoids
from sage.rings.ideal_monoid import IdealMonoid_c
from sage.rings.ideal import Ideal_generic

from sage.rings.integer_ring import ZZ


class IdealMonoid_nc(IdealMonoid_c):
    """
    Base class for the monoid of ideals over a non-commutative ring.

    .. NOTE::

        This class is essentially the same as
        :class:`~sage.rings.ideal_monoid.IdealMonoid_c`,
        but does not complain about non-commutative rings.

    EXAMPLES::

        sage: MS = MatrixSpace(ZZ,2,2)
        sage: MS.ideal_monoid()
        Monoid of ideals of Full MatrixSpace of 2 by 2 dense matrices over Integer Ring
    """
    def __init__(self, R):
        """
        Initialize ``self``.

        INPUT:

        - ``R`` -- a ring

        TESTS::

            sage: from sage.rings.noncommutative_ideals import IdealMonoid_nc
            sage: MS = MatrixSpace(ZZ,2,2)
            sage: IdealMonoid_nc(MS)
            Monoid of ideals of Full MatrixSpace of 2 by 2 dense matrices over Integer Ring
        """
        self._IdealMonoid_c__R = R
        Parent.__init__(self, base=ZZ,
                        category=Monoids())
        self._populate_coercion_lists_()

    def _element_constructor_(self, x):
        r"""
        Create an ideal in this monoid from ``x``.

        INPUT:

        - ``x`` -- an ideal or a list of elements

        TESTS::

            sage: # needs sage.combinat
            sage: A = SteenrodAlgebra(2) # indirect doctest
            sage: IL = A*[A.1+A.2,A.1^2]; IL
            Left Ideal (Sq(2) + Sq(4), Sq(1,1)) of mod 2 Steenrod algebra, milnor basis
            sage: IR = [A.1+A.2,A.1^2]*A; IR
            Right Ideal (Sq(2) + Sq(4), Sq(1,1)) of mod 2 Steenrod algebra, milnor basis
            sage: IT = A*[A.1+A.2,A.1^2]*A; IT
            Twosided Ideal (Sq(2) + Sq(4), Sq(1,1)) of mod 2 Steenrod algebra, milnor basis
            sage: M = IL.parent()
            sage: M([A.0, 1])
            Twosided Ideal (Sq(1), 1) of mod 2 Steenrod algebra, milnor basis

        ::

            sage: # needs sage.combinat
            sage: IL == loads(dumps(IL))
            True
            sage: IR == loads(dumps(IR))
            True
            sage: IT == loads(dumps(IT))
            True
        """
        side = "twosided"
        if isinstance(x, Ideal_nc):
            side = x.side()
            x = x.gens()
        elif isinstance(x, Ideal_generic):
            x = x.gens()
        cdef MonoidElement y = self._IdealMonoid_c__R.ideal(x, side=side)
        y._parent = self
        return y


class Ideal_nc(Ideal_generic):
    """
    Generic non-commutative ideal.

    All fancy stuff such as the computation of Groebner bases must be
    implemented in sub-classes. See :class:`~sage.algebras.letterplace.letterplace_ideal.LetterplaceIdeal`
    for an example.

    EXAMPLES::

        sage: MS = MatrixSpace(QQ,2,2)
        sage: I = MS*[MS.1,MS.2]; I
        Left Ideal
        (
          [0 1]
          [0 0],
        <BLANKLINE>
          [0 0]
          [1 0]
        )
         of Full MatrixSpace of 2 by 2 dense matrices over Rational Field
        sage: [MS.1,MS.2]*MS
        Right Ideal
        (
          [0 1]
          [0 0],
        <BLANKLINE>
          [0 0]
          [1 0]
        )
         of Full MatrixSpace of 2 by 2 dense matrices over Rational Field
        sage: MS*[MS.1,MS.2]*MS
        Twosided Ideal
        (
          [0 1]
          [0 0],
        <BLANKLINE>
          [0 0]
          [1 0]
        )
         of Full MatrixSpace of 2 by 2 dense matrices over Rational Field
    """
    def __init__(self, ring, gens, coerce=True, side='twosided'):
        """
        Initialize ``self``.

        INPUT:

        - ``ring`` -- a ring

        - ``gens`` -- list or tuple of elements

        - ``coerce`` -- boolean (default: ``True``); first coerce the given
          list of elements into the given ring

        - ``side`` -- string (default ``'twosided'``); must be ``'left'``,
          ``'right'`` or ``'twosided'``. Determines whether the ideal is a
          left, right or twosided ideal.

        TESTS::

            sage: MS = MatrixSpace(ZZ,2,2)
            sage: from sage.rings.noncommutative_ideals import Ideal_nc
            sage: Ideal_nc(MS,[MS.1,MS.2], side='left')
            Left Ideal
            (
              [0 1]
              [0 0],
            <BLANKLINE>
              [0 0]
              [1 0]
            )
             of Full MatrixSpace of 2 by 2 dense matrices over Integer Ring
            sage: Ideal_nc(MS,[MS.1,MS.2], side='right')
            Right Ideal
            (
              [0 1]
              [0 0],
            <BLANKLINE>
              [0 0]
              [1 0]
            )
             of Full MatrixSpace of 2 by 2 dense matrices over Integer Ring
        """
        if side not in ['left', 'right', 'twosided']:
            raise ValueError("Ideals are left, right or twosided, but not %s" % side)
        self.__side = side
        Ideal_generic.__init__(self, ring, gens, coerce=coerce)

    def __repr__(self):
        """
        Return a string representation of ``self``.

        TESTS::

            sage: # needs sage.combinat
            sage: A = SteenrodAlgebra(2)
            sage: A*[A.1+A.2,A.1^2]      # indirect doctest
            Left Ideal (Sq(2) + Sq(4), Sq(1,1)) of mod 2 Steenrod algebra, milnor basis
            sage: [A.1+A.2,A.1^2]*A
            Right Ideal (Sq(2) + Sq(4), Sq(1,1)) of mod 2 Steenrod algebra, milnor basis
            sage: A*[A.1+A.2,A.1^2]*A
            Twosided Ideal (Sq(2) + Sq(4), Sq(1,1)) of mod 2 Steenrod algebra, milnor basis
        """
        return "%s Ideal %s of %s" % (self.__side.capitalize(),
                                      self._repr_short(), self.ring())

    def __eq__(self, right):
        """
        Ideals of different sidedness do not compare equal. Apart from
        that, the generators are compared.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: A = SteenrodAlgebra(2)
            sage: IR = [A.1+A.2,A.1^2]*A
            sage: IL = A*[A.1+A.2,A.1^2]
            sage: IT = A*[A.1+A.2,A.1^2]*A
            sage: IT == IL
            False
            sage: IR == [A.1+A.2,A.1^2]*A
            True
        """
        if not isinstance(right, Ideal_nc):
            return False
        if self.side() != right.side():
            return False
        S = set(self.gens())
        T = set(right.gens())
        if S == T:
            return True
        return False

    def __ne__(self, right):
        """
        Ideals of different sidedness do not compare equal. Apart from
        that, the generators are compared.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: A = SteenrodAlgebra(2)
            sage: IR = [A.1+A.2,A.1^2]*A
            sage: IL = A*[A.1+A.2,A.1^2]
            sage: IT = A*[A.1+A.2,A.1^2]*A
            sage: IT != IL
            True
            sage: IR != [A.1+A.2,A.1^2]*A
            False
        """
        return not self.__eq__(right)

    def __hash__(self):
        """
        Return the hash of ``self``.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: A = SteenrodAlgebra(2)
            sage: IR = [A.1+A.2,A.1^2]*A
            sage: IL = A*[A.1+A.2,A.1^2]
            sage: IT = A*[A.1+A.2,A.1^2]*A
            sage: hash(IT) == hash(IL)
            False
            sage: hash(IR) == hash([A.1^2,A.1+A.2]*A)
            True
        """
        return hash((self.parent(), self.__side, frozenset(self.gens())))

    def side(self):
        """
        Return a string that describes the sidedness of this ideal.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: A = SteenrodAlgebra(2)
            sage: IL = A*[A.1+A.2,A.1^2]
            sage: IR = [A.1+A.2,A.1^2]*A
            sage: IT = A*[A.1+A.2,A.1^2]*A
            sage: IL.side()
            'left'
            sage: IR.side()
            'right'
            sage: IT.side()
            'twosided'
        """
        return self.__side

    def __mul__(self, other):
        r"""
        Multiply ``self`` with ``other``.

        Multiplication of a one-sided ideal with its ring from the other side
        yields a two-sided ideal.

        Let `L` (resp. `R`) be a left (resp. right) ideal, then the product
        `LR` is a twosided ideal generated by `x y`, where `x` (resp. `y`)
        is a generator of `L` (resp. `R`).

        .. TODO::

            The product of left (resp. right) ideals is a left
            (resp. right) ideal. However, these do not necessarily have
            simple generating sets.

        TESTS::

            sage: MS = MatrixSpace(QQ,2,2)
            sage: IL = MS * [2*MS.0,3*MS.1]; IL
            Left Ideal
            (
              [2 0]
              [0 0],
            <BLANKLINE>
              [0 3]
              [0 0]
            )
             of Full MatrixSpace of 2 by 2 dense matrices over Rational Field
            sage: IR = MS.3*MS; IR
            Right Ideal
            (
              [0 0]
              [0 1]
            )
             of Full MatrixSpace of 2 by 2 dense matrices over Rational Field
            sage: IL * MS
            Twosided Ideal
            (
              [2 0]
              [0 0],
            <BLANKLINE>
              [0 3]
              [0 0]
            )
             of Full MatrixSpace of 2 by 2 dense matrices over Rational Field

            sage: IL * IR
            Twosided Ideal
            (
              [0 3]
              [0 0]
            )
             of Full MatrixSpace of 2 by 2 dense matrices over Rational Field

            sage: IR * IR
            Traceback (most recent call last):
            ...
            NotImplementedError: cannot multiply non-commutative ideals
        """
        if isinstance(other, Ideal_nc) and self.ring() == other.ring():
            if self.side() == "left" and other.side() == "right":
                gens = [z for z in (x * y for x in self.gens() for y in other.gens()) if z]
                return self.ring().ideal(gens, side='twosided')
            raise NotImplementedError("cannot multiply non-commutative ideals")

        if not isinstance(other, Ideal_nc):
            # Perhaps other is a ring and thus has its own
            # multiplication.
            if other == self.ring():
                if self.side() == 'right':
                    return self
                return self.ring().ideal(self.gens(), side='twosided')
        if not isinstance(self, Ideal_nc):
            # This may happen...
            if self == other.ring():
                if other.side() == 'left':
                    return other
                return other.ring().ideal(other.gens(), side='twosided')

        raise NotImplementedError("cannot multiply non-commutative ideals")



# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.symbolic
r"""
Tropical Varieties

A tropical variety is a piecewise-linear geometric object derived from
a classical algebraic variety by using tropical mathematics, where the
tropical semiring replaces the usual arithmetic operations.

AUTHORS:

- Verrel Rievaldo Wijaya (2024-06): initial version

REFERENCES:

- [Bru2014]_
- [Mac2015]_
- [Fil2017]_
"""

# ****************************************************************************
#       Copyright (C) 2024 Verrel Rievaldo Wijaya <verrelrievaldo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.sage_object import SageObject
from sage.structure.unique_representation import UniqueRepresentation

from sage.rings.rational_field import QQ
from sage.rings.infinity import infinity

class TropicalVariety(UniqueRepresentation, SageObject):
    r"""
    A tropical variety in `\RR^n`.

    A tropical variety is defined as a corner locus of tropical polynomial
    function. This means it consist of all points in `\RR^n` for which
    the minimum (maximum) of the function is attained at least twice.

    We represent the tropical variety as a list of lists, where the
    inner list consist of three parts. The first one is a parametric
    equations for tropical roots. The second one is the condition
    for parameters. The third one is the order of the corresponding
    component.

    INPUT:

    - ``poly`` -- a :class:`TropicalMPolynomial`

    ALGORITHM:

    We need to determine a corner locus of this tropical polynomial
    function, which is all points `(x_1, x_2, \ldots, x_n)` for which
    the maximum (minimum) is obtained at least twice. First, we convert
    each monomial to its corresponding linear function. Then for each two
    monomials of polynomial, we find the points where their values are
    equal. Since we attempt to solve the equality of two equations in `n`
    variables, the solution set will be described by `n-1` parameters.

    Next, we need to check if the value of previous two monomials at the
    points in solution set is really the maximum (minimum) of function.
    We do this by solving the inequality of the previous monomial with all
    other monomials in the polynomial after substituting the parameter.
    This will give us the condition of parameters. Each of this condition
    is then combined by union operator. If this final condition is not an
    empty set, then it represent one component of tropical root. Then we
    calculate the weight of this particular component by the maximum of
    gcd of the numbers `|i-k|` and `|j-l|` for all pairs `(i,j)` and
    `(k,l)` such that the value of on this component is given by the
    corresponding monomials.

    EXAMPLES:

    We construct a tropical variety in `\RR^2`, where it is called a
    tropical curve::

        sage: T = TropicalSemiring(QQ, use_min=False)
        sage: R.<x,y> = PolynomialRing(T)
        sage: p1 = R(1)*x + x*y + R(0); p1
        0*x*y + 1*x + 0
        sage: tv = p1.tropical_variety(); tv
        Tropical curve of 0*x*y + 1*x + 0
        sage: tv.components()
        [[(t1, 1), [t1 >= -1], 1], [(-1, t1), [t1 <= 1], 1], [(-t1, t1), [t1 >= 1], 1]]
        sage: tv.vertices()
        {(-1, 1)}
        sage: tv.plot()                                                                 # needs sage.plot
        Graphics object consisting of 3 graphics primitives

    .. PLOT::
        :width: 300 px

        T = TropicalSemiring(QQ, use_min=False)
        R = PolynomialRing(T, ('x,y'))
        x, y = R.gen(), R.gen(1)
        p1 = R(1)*x + x*y + R(0)
        sphinx_plot(p1.tropical_variety().plot())

    A slightly different result will be obtained if we use min-plus algebra
    for the base tropical semiring::

        sage: T = TropicalSemiring(QQ, use_min=True)
        sage: R.<x,y> = PolynomialRing(T)
        sage: p1 = R(1)*x + x*y + R(0)
        sage: tv = p1.tropical_variety(); tv
        Tropical curve of 0*x*y + 1*x + 0
        sage: tv.components()
        [[(t1, 1), [t1 <= -1], 1], [(-1, t1), [t1 >= 1], 1], [(-t1, t1), [t1 <= 1], 1]]
        sage: tv.plot()                                                                 # needs sage.plot
        Graphics object consisting of 3 graphics primitives

    .. PLOT::
        :width: 300 px

        T = TropicalSemiring(QQ, use_min=True)
        R = PolynomialRing(T, ('x,y'))
        x, y = R.gen(), R.gen(1)
        p1 = R(1)*x + x*y + R(0)
        sphinx_plot(p1.tropical_variety().plot())

    Tropical variety can consist of multiple components with varying orders::

        sage: T = TropicalSemiring(QQ, use_min=False)
        sage: R.<x,y> = PolynomialRing(T)
        sage: p1 = R(7) + T(4)*x + y + R(4)*x*y + R(3)*y^2 + R(-3)*x^2
        sage: tv = p1.tropical_variety(); tv
        Tropical curve of (-3)*x^2 + 4*x*y + 3*y^2 + 4*x + 0*y + 7
        sage: tv.components()
        [[(3, t1), [t1 <= 0], 1],
        [(-t1 + 3, t1), [0 <= t1, t1 <= 2], 1],
        [(t1, 2), [t1 <= 1], 2],
        [(t1, 0), [3 <= t1, t1 <= 7], 1],
        [(7, t1), [t1 <= 0], 1],
        [(t1 - 1, t1), [2 <= t1], 1],
        [(t1 + 7, t1), [0 <= t1], 1]]
        sage: tv.plot()                                                                 # needs sage.plot
        Graphics object consisting of 8 graphics primitives

    .. PLOT::
        :width: 300 px

        T = TropicalSemiring(QQ, use_min=False)
        R = PolynomialRing(T, ('x,y'))
        x, y = R.gen(), R.gen(1)
        p1 = R(7) + T(4)*x + y + R(4)*x*y + R(3)*y**2 + R(-3)*x**2
        sphinx_plot(p1.tropical_variety().plot())

    If the tropical polynomial have `n>2` variables, then the result will be
    a tropical hypersurface embedded in a real space `\RR^n`::

        sage: T = TropicalSemiring(QQ)
        sage: R.<w,x,y,z> = PolynomialRing(T)
        sage: p1 = x*y + R(-1/2)*x*z + R(4)*z^2 + w*x
        sage: tv = p1.tropical_variety(); tv
        Tropical hypersurface of 0*w*x + 0*x*y + (-1/2)*x*z + 4*z^2
        sage: tv.components()
        [[(t1, t2, t3 - 1/2, t3), [t2 - 9/2 <= t3, t3 <= t1 + 1/2, t2 - 5 <= t1], 1],
        [(t1, 2*t2 - t3 + 4, t3, t2), [t3 + 1/2 <= t2, t3 <= t1], 1],
        [(t1, t2, t1, t3), [max(t1 + 1/2, 1/2*t1 + 1/2*t2 - 2) <= t3], 1],
        [(t1, t2 + 9/2, t3, t2), [t2 <= min(t3 + 1/2, t1 + 1/2)], 1],
        [(t1 - 1/2, t2, t3, t1), [t2 - 9/2 <= t1, t1 <= t3 + 1/2, t2 - 5 <= t3], 1],
        [(2*t1 - t2 + 4, t2, t3, t1), [t1 <= min(1/2*t2 + 1/2*t3 - 2, t2 - 9/2)], 1]]
    """
    def __init__(self, poly):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: tv = (x+y).tropical_variety()
            sage: TestSuite(tv).run()

        TESTS::

            sage: from sage.rings.semirings.tropical_variety import TropicalVariety
            sage: R.<x,y> = QQ[]
            sage: p1 = x + y
            sage: TropicalVariety(p1)
            Traceback (most recent call last):
            ...
            ValueError: x + y is not a multivariate tropical polynomial
        """
        import operator
        from itertools import combinations
        from sage.symbolic.ring import SR
        from sage.symbolic.relation import solve
        from sage.arith.misc import gcd
        from sage.rings.semirings.tropical_mpolynomial import TropicalMPolynomial

        if not isinstance(poly, TropicalMPolynomial):
            raise ValueError(f"{poly} is not a multivariate tropical polynomial")

        self._poly = poly
        self._hypersurface = []
        tropical_roots = []
        variables = [SR.var(name)
                     for name in poly.parent().variable_names()]

        # Convert each term to its linear function
        linear_eq = {}
        pd = poly.monomial_coefficients()
        for key in pd:
            eq = sum(variables[i] * e for i, e in enumerate(key))
            eq += pd[key].lift()
            linear_eq[key] = eq
        temp_keys = []
        temp_order = []

        # Checking for all possible combinations of two terms
        for keys in combinations(pd, 2):
            sol = solve(linear_eq[keys[0]] == linear_eq[keys[1]], variables)

            # Parametric solution of the chosen two terms
            final_sol = [s.right() for s in sol[0]]
            xy_interval = []
            xy_interval.append(tuple(final_sol))

            # Comparing with other terms
            min_max = linear_eq[keys[0]]
            for i, v in enumerate(variables):
                min_max = min_max.subs(**{str(v): final_sol[i]})
            all_sol_compare = []
            no_solution = False
            for compare in pd:
                if compare not in keys:
                    temp_compare = linear_eq[compare]
                    for i, v in enumerate(variables):
                        temp_compare = temp_compare.subs(**{str(v): final_sol[i]})
                    if min_max == temp_compare:
                        sol_compare = [[]]
                    elif poly.parent().base()._use_min:
                        sol_compare = solve(min_max < temp_compare, variables)
                    else:
                        sol_compare = solve(min_max > temp_compare, variables)
                    if sol_compare:
                        if isinstance(sol_compare[0], list):
                            if sol_compare[0]:
                                all_sol_compare.append(sol_compare[0][0])
                        else:  # solution is unbounded on one side
                            all_sol_compare.append(sol_compare[0])
                    else:
                        no_solution = True
                        break

            # Solve the condition for parameter
            if not no_solution:
                parameter = set()
                for sol in all_sol_compare:
                    parameter = parameter.union(set(sol.variables()))
                parameter_solution = solve(all_sol_compare, list(parameter))
                if parameter_solution:
                    xy_interval.append(parameter_solution[0])
                    tropical_roots.append(xy_interval)
                    # Calculate the order
                    index_diff = [abs(ai - bi)
                                  for ai, bi in zip(keys[0], keys[1])]
                    order = gcd(index_diff)
                    temp_order.append(order)
                    temp_keys.append(keys)

        # Changing all the operator's symbol to <= or >=
        self._keys = []
        components = []
        dim_param = 0
        if tropical_roots:
            dim_param = len(tropical_roots[0][0]) - 1
        vars = [SR.var(f't{i}') for i in range(1, dim_param + 1)]
        for arg in tropical_roots:
            subs_dict = {}
            index_vars = 0
            new_eq = []
            for eq in arg[0]:
                var_eq = eq.variables()
                for var in var_eq:
                    if var not in subs_dict:
                        subs_dict[var] = vars[index_vars]
                        index_vars += 1
                new_eq.append(eq.subs(subs_dict))
            new_eq = tuple(new_eq)
            arg.remove(arg[0])
            arg.insert(0, new_eq)
            if not arg[1] or not isinstance(arg[1], list):
                arg[1] = []
                for var in vars:
                    expr1 = -infinity < var
                    expr2 = var < infinity
                    arg[1].append(expr1)
                    arg[1].append(expr2)
            else:
                params = arg[1]
                arg.remove(params)
                new_param = []
                for param in params:
                    lhs = param.lhs().subs(subs_dict)
                    rhs = param.rhs().subs(subs_dict)
                    if param.operator() == operator.gt:
                        expr = lhs >= rhs
                    else:
                        expr = lhs <= rhs
                    new_param.append(expr)
                arg.insert(1, new_param)
            components.append(arg)

        # Determine the order of each component
        self._vars = vars
        final_order = []
        for i, component in enumerate(components):
            if component not in self._hypersurface:
                self._hypersurface.append(component)
                final_order.append(temp_order[i])
                self._keys.append(temp_keys[i])
            else:
                index = self._hypersurface.index(component)
                if temp_order[i] > final_order[index]:
                    final_order[index] = temp_order[i]
                    self._keys[index] = temp_keys[i]
        for i in range(len(self._hypersurface)):
            self._hypersurface[i].append(final_order[i])

    def dimension(self):
        """
        Return the dimension of ``self``.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<a,x,y,z> = PolynomialRing(T)
            sage: p1 = x*y + R(-1)*x*z
            sage: p1.tropical_variety().dimension()
            4
        """
        return self._poly.parent().ngens()

    def number_of_components(self):
        """
        Return the number of components that make up ``self``.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<a,x,y,z> = PolynomialRing(T)
            sage: p1 = x*y*a + x*z + y^2 + a*x + y + z
            sage: p1.tropical_variety().number_of_components()
            13
        """
        from sage.rings.integer_ring import ZZ
        return ZZ(len(self._hypersurface))

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<w,x,y,z> = PolynomialRing(T)
            sage: (w).tropical_variety()
            Tropical hypersurface of 0*w
        """
        return f"Tropical hypersurface of {self._poly}"

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<w,x,y,z> = PolynomialRing(T)
            sage: tv = (R(1)*w^2 + x*y*z + R(-1)).tropical_variety()
            sage: latex(tv)
            TV\left(0 x y z + 1 w^{2} + \left(-1\right)\right)
        """
        return f"TV\\left({self._poly._latex_()}\\right)"

    def components(self):
        """
        Return all components of ``self``.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<a,x,y,z> = PolynomialRing(T)
            sage: tv = (a+x+y+z).tropical_variety()
            sage: tv.components()
            [[(t1, t1, t2, t3), [t1 <= min(t3, t2)], 1],
             [(t1, t2, t1, t3), [t1 <= t3, t1 <= t2], 1],
             [(t1, t2, t3, t1), [t1 <= min(t3, t2)], 1],
             [(t1, t2, t2, t3), [t2 <= t3, t2 <= t1], 1],
             [(t1, t2, t3, t2), [t2 <= min(t3, t1)], 1],
             [(t1, t2, t3, t3), [t3 <= min(t1, t2)], 1]]
        """
        return self._hypersurface

    def _components_intersection(self):
        r"""
        Return the intersection of three or more components of ``self``.

        For a tropical variety in `\RR^n`, the intersection is characterized
        by a linear equation in `\RR^{n-1}`. Specifically, this becomes a
        vertex for tropical curve and an edges for tropical surface.

        OUTPUT:

        A dictionary where the keys represent component indices and the
        values are lists of tuples. Each tuple contains a parametric
        equation of points and the corresponding parameter's condition.

        EXAMPLES:

        In two dimension, it will provide vertices that are incident with
        each component::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = R(2)*x^2 + x*y + R(2)*y^2 + x + R(-1)*y + R(3)
            sage: tv = p1.tropical_variety()
            sage: tv._components_intersection()
            {0: [((-2, 0), {})],
            1: [((-2, 0), {})],
            2: [((-1, -3), {})],
            3: [((-2, 0), {}), ((-1, 0), {})],
            4: [((-1, -3), {}), ((-1, 0), {})],
            5: [((-1, -3), {})],
            6: [((-1, 0), {}), ((3, 4), {})],
            7: [((3, 4), {})],
            8: [((3, 4), {})]}

        In three dimensions, it will provide all parametric equations of
        lines that lie within each component::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y,z> = PolynomialRing(T)
            sage: p1 = x + y + z + x^2
            sage: tv = p1.tropical_variety()
            sage: tv._components_intersection()
            {0: [((t2, t2, t2), {0 <= t2}), ((0, 0, t2), {0 <= t2})],
            1: [((0, t2, 0), {0 <= t2}), ((t2, t2, t2), {0 <= t2})],
            2: [((0, t1, 0), {0 <= t1}), ((0, 0, t2), {0 <= t2})],
            3: [((t1, t1, t1), {0 <= t1}), ((t1, 2*t1, 2*t1), {t1 <= 0})],
            4: [((1/2*t2, t2, t2), {t2 <= 0}), ((0, 0, t2), {0 <= t2})],
            5: [((0, t2, 0), {0 <= t2}), ((1/2*t2, t2, t2), {t2 <= 0})]}
        """
        import operator
        from sage.functions.min_max import max_symbolic, min_symbolic
        from sage.symbolic.relation import solve
        from sage.sets.set import Set

        def update_result(result):
            sol_param = solve(new_expr, vars)
            sol_param_sim = set()
            for sol in sol_param:
                if sol == []:
                    for v in vars:
                        if v != var:
                            sol_param_sim.add(v < infinity)
                elif isinstance(sol, list):
                    for eqn in sol:
                        if eqn.operator() == operator.eq:
                            if not eqn.rhs().is_numeric():
                                eqn_var = eqn.rhs().variables()
                                param_var = [v for v in eqn_var if v in vars]
                                if not param_var:
                                    v = eqn.lhs()
                                    if v != var:
                                        sol_param_sim.add(v < infinity)
                        elif eqn.operator() == operator.lt:
                            sol_param_sim.add(eqn.lhs() <= eqn.rhs())
                        elif eqn.operator() == operator.gt:
                            sol_param_sim.add(eqn.lhs() >= eqn.rhs())
                else:
                    sol_param_sim.add(sol)

            # Checking there are no conditions with the same variables
            # that use the <= and >= operators simultaneously
            unique_sol_param = set()
            temp = list(sol_param_sim)
            op_temp = {i: set(temp[i].operands()) for i in range(len(temp))}
            for s_value in op_temp.values():
                match_keys = [k for k, v in op_temp.items() if v == s_value]
                if len(match_keys) == 1:
                    for i in match_keys:
                        unique_sol_param.add(temp[i])

            if (unique_sol_param) or (self.dimension() == 2):
                if not unique_sol_param:
                    unique_sol_param = Set()
                if index not in result:
                    result[index] = [(tuple(points), unique_sol_param)]
                else:
                    result[index].append((tuple(points), unique_sol_param))

        result = {}
        vars = self._vars
        for index, comp in enumerate(self._hypersurface):
            for expr in comp[1]:
                left = expr.lhs()
                right = expr.rhs()
                # If the lhs contains a min or max operator
                if (left.operator() == max_symbolic) or (left.operator() == min_symbolic):
                    for operand in expr.lhs().operands():
                        points = list(comp[0])
                        new_expr = [e.subs(**{str(right): operand}) for e in comp[1]]
                        for i, p in enumerate(points):
                            new_eq = p.subs(**{str(right): operand})
                            points[i] = new_eq
                        update_result(result)
                # If the rhs contains a min or max operator
                elif (right.operator() == max_symbolic) or (right.operator() == min_symbolic):
                    for operand in expr.rhs().operands():
                        points = list(comp[0])
                        new_expr = [e.subs(**{str(left): operand}) for e in comp[1]]
                        for i, p in enumerate(points):
                            new_eq = p.subs(**{str(left): operand})
                            points[i] = new_eq
                        update_result(result)
                else:
                    var = expr.variables()[0]
                    points = list(comp[0])
                    subs_expr = solve(left == right, var)[0].rhs()
                    new_expr = [e.subs(**{str(var): subs_expr}) for e in comp[1]]
                    for i, p in enumerate(points):
                        new_eq = p.subs(**{str(var): subs_expr})
                        points[i] = new_eq
                    update_result(result)
        return result

    def weight_vectors(self):
        r"""
        Return the weight vectors for each unique intesection of
        components of ``self``.

        Weight vectors are a list of vectors associated with each
        unique intersection of the components of tropical variety.
        Each vector is a normal vector to a component with respect
        to the unique intersection lying within that component.

        Assume ``self`` is a `n`-dimensional tropical variety.
        Suppose `L` is an intersection lying within the components
        `S_1, \ldots, S_k` with respective weights `w_1, \ldots, w_k`.
        This `L` is a linear structure in `\RR^{n-1}` and has `n-1`
        direction vectors `d_1,d_2,\ldots, d_{n-1}`. Each component
        `S_1, \ldots, S_k` has a normal vector `n_1, \ldots, n_k`.
        Then, we scale each normal vector to an integer vector such
        that the greatest common divisor of its elements is 1.

        The weight vector of a component `S_i` with respect to `L`
        can be found by calculating the cross product between direction
        vectors of `L` and normal vector `n_i`.These vectors will
        satisfy the balancing condition `\sum_{i=1}^k w_k v_k = 0`.

        OUTPUT:

        A tuple of three dictionaries. The first dictionary contains
        equations representing the intersections. The second dictionary
        contains indices of components that contains the intersection.
        The third dictionary contains lists of vectors.

        EXAMPLES:

        Weight vectors of tropical surface::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y,z> = PolynomialRing(T)
            sage: p = x^2 + R(-1)*y + z + R(1)
            sage: tv = p.tropical_variety()
            sage: tv.weight_vectors()
            ({0: ((1/2*u2, u2 + 1, u2), {u2 <= 1}),
              1: ((1/2, 2, u2), {1 <= u2}),
              2: ((1/2, u2, 1), {2 <= u2}),
              3: ((u1, 2, 1), {(1/2) <= u1})},
             {0: [0, 1, 3], 1: [0, 2, 4], 2: [1, 2, 5], 3: [3, 4, 5]},
             {0: [(1, 2, -5/2), (1, -5/2, 2), (-2, 1/2, 1/2)],
              1: [(-1, -2, 0), (0, 2, 0), (1, 0, 0)],
              2: [(1, 0, 2), (0, 0, -2), (-1, 0, 0)],
              3: [(0, 1, 1), (0, 0, -1), (0, -1, 0)]})

        TESTS:

        Checking the balance condition of weight vectors::

            sage: T = TropicalSemiring(QQ)
            sage: R.<a,b,c,d> = PolynomialRing(T)
            sage: f = R.random_element()
            sage: vec = f.tropical_variety().weight_vectors()[2].values()
            sage: all(a == vector([0,0,0,0]) for a in [sum(lst) for lst in vec])
            True
        """
        from sage.symbolic.ring import SR
        from sage.symbolic.relation import solve
        from sage.calculus.functional import diff
        from sage.arith.misc import gcd
        from sage.matrix.constructor import matrix
        from sage.modules.free_module_element import vector, zero_vector
        from itertools import combinations

        dim = self.dimension()
        t = SR.var('t')
        t_vars = [SR.var('t{}'.format(i)) for i in range(dim)]
        u_vars = [SR.var('u{}'.format(i)) for i in range(dim)]
        convert_tu = {ti: ui for ti, ui in zip(t_vars, u_vars)}
        CI = self._components_intersection()
        unique_line = set()
        index_line = {}
        line_comps = {}
        index = 0

        # Find the unique intersection between multiple components and
        # the indices of the components containing this intersection.
        for i, lines in CI.items():
            for line in lines:
                eqn = line[0]
                is_unique = True
                for uniq in unique_line:
                    subs_index = -1
                    for j in range(dim):
                        if eqn[j] != uniq[j]:
                            subs_index = j
                            break
                    if subs_index == -1:
                        new_line = eqn
                        is_unique = False
                        break
                    subs_dict = {}
                    while len(subs_dict) != dim-2 and subs_index < dim:
                        eq1 = eqn[subs_index].subs(subs_dict)
                        vib = None
                        for unk in eq1.variables():
                            if unk not in subs_dict:
                                if unk in t_vars:
                                    vib = unk
                                    break
                        if vib:
                            eq1 = eq1.subs(**{str(vib): t})
                            eq2 = uniq[subs_index]
                            temp_sol = solve(eq1 == eq2, t)
                            if temp_sol:
                                temp_sol = temp_sol[0].rhs()
                                if not temp_sol.is_numeric():
                                    subs_dict[vib] = temp_sol
                        subs_index += 1
                    if subs_dict:
                        new_line = []
                        for l in eqn:
                            for key, value in subs_dict.items():
                                l = l.subs(key == value)
                            new_line.append(l)
                        if tuple(new_line) in unique_line:
                            is_unique = False
                            break
                if is_unique:
                    new_eqn = tuple([eq.subs(convert_tu) for eq in eqn])
                    cdns = line[1]
                    new_cdn = set([cdn.subs(convert_tu) for cdn in cdns])
                    unique_line.add(new_eqn)
                    index_line[index] = tuple([new_eqn, new_cdn])
                    line_comps[index] = [i]
                    index += 1
                else:
                    match_key = [k for k, v in index_line.items() if v[0] == tuple(new_line)][0]
                    line_comps[match_key].append(i)

        WV = {i: [] for i in range(len(line_comps)) if len(line_comps[i]) > 1}
        for k in WV:
            # Calculate direction vector of the line
            dir_vecs = []
            line = index_line[k][0]
            all_var = set()
            for l in line:
                for v in l.variables():
                    all_var.add(v)
            for vpar in all_var:
                par_drv = []
                for l in line:
                    par_drv.append(QQ(diff(l, vpar)))
                par_drv = vector(par_drv)
                dir_vecs.append(par_drv)

            # Calculate the outgoing normal vector of each surface in the
            # direction of the line
            for i in line_comps[k]:
                surface = self._hypersurface[i][0]
                drv_vectors = []
                for vpar in self._vars:
                    temp_vec = []
                    for s in surface:
                        temp_vec.append(QQ(diff(s, vpar)))
                    temp_vec = vector(temp_vec)
                    drv_vectors.append(temp_vec)
                temp = [t_vars]
                for vec in drv_vectors:
                    temp.append(vec)
                vec_matrix = matrix(SR, temp)
                normal_vec = vec_matrix.det()
                temp_nor = []
                for tvar in t_vars:
                    temp_nor.append(QQ(diff(normal_vec, tvar)))
                normal_vec = vector(temp_nor)
                normal_vec *= 1/gcd(normal_vec)

                # Calculate the weight vector
                temp_final = [t_vars]
                for v in dir_vecs:
                    temp_final.append(v)
                temp_final.append(normal_vec)
                vec_matrix = matrix(SR, temp_final)
                weight_vec = vec_matrix.det()
                temp_weight = []
                for tvar in t_vars:
                    temp_weight.append(QQ(diff(weight_vec, tvar)))
                weight_vec = vector(temp_weight)
                order = self._hypersurface[i][2]
                weight_vec *= order
                WV[k].append(weight_vec)

            balance = False
            for i in range(1, len(WV[k])+1):
                for j in combinations(range(len(WV[k])), i):
                    test_vectors = [v for v in WV[k]]
                    for idx in j:
                        test_vectors[idx] = -test_vectors[idx]
                    if sum(test_vectors) == zero_vector(QQ, dim):
                        WV[k] = test_vectors
                        balance = True
                        break
                if balance:
                    break

        return index_line, line_comps, WV


class TropicalSurface(TropicalVariety):
    r"""
    A tropical surface in `\RR^3`.

    The tropical surface consists of planar regions and facets, which we
    can call cells. These cells are connected in such a way that they form
    a piecewise linear structure embedded in three-dimensional space. These
    cells meet along edges, where the balancing condition is satisfied.
    This balancing condition ensures that the sum of the outgoing normal
    vectors at each edge is zero, reflecting the equilibrium.

    EXAMPLES::

        sage: T = TropicalSemiring(QQ, use_min=False)
        sage: R.<x,y,z> = PolynomialRing(T)
        sage: p1 = x + y + z + R(0)
        sage: tv = p1.tropical_variety(); tv
        Tropical surface of 0*x + 0*y + 0*z + 0
        sage: tv.components()
        [[(t1, t1, t2), [t2 <= t1, 0 <= t1], 1],
        [(t1, t2, t1), [max(0, t2) <= t1], 1],
        [(0, t1, t2), [t2 <= 0, t1 <= 0], 1],
        [(t1, t2, t2), [max(0, t1) <= t2], 1],
        [(t1, 0, t2), [t2 <= 0, t1 <= 0], 1],
        [(t1, t2, 0), [t1 <= 0, t2 <= 0], 1]]
    """
    def _axes(self):
        r"""
        Set the default axes for ``self``.

        This default axes is used for the 3d plot. The axes is centered
        around where the intersection of the components occured so it
        gives a nice visual representation for the interactions between
        different components of the surface. Additionally, it enhances
        the visibility and interpretation of how the components align
        and interact in three-dimensional space.

        OUTPUT:

        A list of three lists, where the first inner list represent value
        of x-axis, the second inner list represent value of y-axis, and
        the third inner list represent value of z-axis. If there are
        either no components or only one component, the axis will be set
        to `[[-1, 1], [-1, 1], [-1, 1]]`.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y,z> = PolynomialRing(T)
            sage: p1 = x
            sage: p1.tropical_variety()._axes()
            [[-1, 1], [-1, 1], [-1, 1]]
            sage: p2 = x + y + z + x^2 + R(1)
            sage: p2.tropical_variety()._axes()
            [[-1, 2], [-1, 2], [-1, 2]]
        """
        from sage.symbolic.relation import solve
        from sage.arith.srange import srange

        if not self._hypersurface:
            return [[-1, 1], [-1, 1], [-1, 1]]
        elif len(self._hypersurface) == 1:
            bound = 1
            for eqn in self._hypersurface[0][0]:
                for op in eqn.operands():
                    if op.is_numeric():
                        bound = max(op, bound)
            return [[-bound, bound]] * 3

        u_set = set()
        v_set = set()
        for comp in self._hypersurface:
            list_expr = []
            temp_u = set()
            temp_v = set()
            for expr in comp[1]:
                if expr.lhs().is_numeric():
                    if bool(expr.rhs() == self._vars[0]):
                        temp_u.add(expr.lhs())
                    else:
                        temp_v.add(expr.lhs())
                elif expr.rhs().is_numeric():
                    if bool(expr.lhs() == self._vars[0]):
                        temp_u.add(expr.rhs())
                    else:
                        temp_v.add(expr.rhs())
                else:
                    list_expr.append(expr)
            if not temp_u:
                temp_u.add(0)
            if not temp_v:
                temp_v.add(0)
            for expr in list_expr:
                for u in temp_u:
                    sol = solve(expr.subs(**{str(self._vars[0]): u}), self._vars[1])
                    if not sol:
                        temp_v.add(0)
                    elif not sol[0]:
                        temp_v.add(0)
                    else:
                        temp_v.add(sol[0][0].rhs())
                for v in temp_v:
                    sol = solve(expr.subs(**{str(self._vars[1]): v}), self._vars[0])
                    if not sol:
                        temp_u.add(0)
                    elif not sol[0]:
                        temp_u.add(0)
                    else:
                        temp_u.add(sol[0][0].rhs())
            u_set = u_set.union(temp_u)
            v_set = v_set.union(temp_v)
        axes = [[min(u_set)-1, max(u_set)+1], [min(v_set)-1, max(v_set)+1]]

        # Calculate the z-axis
        step = 10
        du = (axes[0][1]-axes[0][0]) / step
        dv = (axes[1][1]-axes[1][0]) / step
        u_range = srange(axes[0][0], axes[0][1]+du, du)
        v_range = srange(axes[1][0], axes[1][1]+dv, dv)
        zmin, zmax = None, None
        for comp in self._hypersurface:
            for u in u_range:
                for v in v_range:
                    checkpoint = True
                    for exp in comp[1]:
                        final_exp = exp.subs(**{str(self._vars[0]): u, str(self._vars[1]): v})
                        if not final_exp:
                            checkpoint = False
                            break
                    if checkpoint:
                        z = comp[0][2].subs(**{str(self._vars[0]): u, str(self._vars[1]): v})
                        if (zmin is None) and (zmax is None):
                            zmin = z
                            zmax = z
                        else:
                            zmin = min(z, zmin)
                            zmax = max(z, zmax)
        axes.append([zmin, zmax])
        return axes

    def _polygon_vertices(self):
        r"""
        Return the vertices of the polygon for each components of ``self``
        to be used for plotting.

        OUTPUT:

        A dictionary where the keys represent component indices and the
        values are a set of points in three dimensional space.

        EXAMPLES:

        A tropical surface with only one component::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y,z> = PolynomialRing(T)
            sage: p1 = x + z
            sage: tv1 = p1.tropical_variety()
            sage: tv1._polygon_vertices()
            {0: {(-1, -1, -1), (-1, 1, -1), (1, -1, 1), (1, 1, 1)}}

        A tropical surface with multiple components::

            sage: p2 = x^2 + x + y + z + R(1)
            sage: tv2 = p2.tropical_variety()
            sage: tv2._polygon_vertices() # long time
            {0: {(0, 0, 0), (0, 0, 2), (1, 1, 1), (2, 2, 2)},
            1: {(0, 0, 0), (0, 2, 0), (1, 1, 1), (2, 2, 2)},
            2: {(0, 0, 0), (0, 0, 2), (0, 2, 0), (0, 2, 2)},
            3: {(1, 1, 1), (1, 1, 2), (1, 2, 1), (1, 2, 2)},
            4: {(-1/2, -1, -1), (0, 0, 0), (1, 1, 1), (2, -1, -1), (2, 1, 1)},
            5: {(-1/2, -1, -1), (-1/2, -1, 2), (0, 0, 0), (0, 0, 2)},
            6: {(1, 1, 1), (1, 1, 2), (2, 1, 1), (2, 1, 2)},
            7: {(-1/2, -1, -1), (-1/2, 2, -1), (0, 0, 0), (0, 2, 0)},
            8: {(1, 1, 1), (1, 2, 1), (2, 1, 1), (2, 2, 1)}}
        """
        from sage.symbolic.relation import solve
        from sage.sets.real_set import RealSet

        poly_verts = {i: set() for i in range(self.number_of_components())}
        axes = self._axes()
        comps = self.components()
        vars = self._vars
        comps_int = self._components_intersection()

        # Find the inside vertices (intersection of components)
        for index, lines in comps_int.items():
            for line in lines:
                v = list(line[1])[0].variables()[0]
                for param in line[1]:
                    left = param.lhs()
                    right = param.rhs()
                    if left.is_numeric():
                        vertex = [QQ(e.subs(**{str(v): left})) for e in line[0]]
                        poly_verts[index].add(tuple(vertex))
                    elif right.is_numeric():
                        vertex = [QQ(e.subs(**{str(v): right})) for e in line[0]]
                        poly_verts[index].add(tuple(vertex))

        def find_edge_vertices(i):
            j = (i+1) % 2
            if i == 0:  # interval for t1
                interval = interval1
            else:  # interval for t2
                interval = interval2
            for p in [interval.inf(), interval.sup()]:
                new_param = [e.subs(**{str(vars[i]): p}) for e in comps[index][1]]
                sol = solve(new_param, vars[j])
                if sol:
                    interval_param = RealSet()
                    for s in sol:
                        if s != []:
                            # Handle cases where 's' is not a list (s = r1 < +Infinity),
                            # or if 's' is a list, ensure that its elements define a real
                            # interval (catch invalid cases like s = [t1 == r100]).
                            try:
                                interval_param += RealSet(s[0])
                            except (IndexError, ValueError):
                                interval_param += RealSet(-infinity, infinity)
                        else:
                            interval_param += RealSet(-infinity, infinity)
                    interval_param = interval_param.intersection(interval2)
                    if is_doublevar:
                        int1 = RealSet()
                        for s1 in sol1:
                            subs1 = solve(s1[0].subs(**{str(vars[i]): p}), vars[j])
                            try:
                                int1 += RealSet(subs1[0])
                            except TypeError:
                                int1 += RealSet(subs1[0][0])
                        int2 = RealSet()
                        for s2 in sol2:
                            subs2 = solve(s2[0].subs(**{str(vars[i]): p}), vars[j])
                            try:
                                int2 += RealSet(subs2[0])
                            except TypeError:
                                int2 += RealSet(subs2[0][0])
                        final_int = int1.intersection(int2)
                        interval_param = interval_param.intersection(final_int)
                    if interval_param:
                        vertex1 = [QQ(e.subs(**{str(vars[i]): p, str(vars[j]): interval_param.inf()})) for e in comps[index][0]]
                        vertex2 = [QQ(e.subs(**{str(vars[i]): p, str(vars[j]): interval_param.sup()})) for e in comps[index][0]]
                        poly_verts[index].add(tuple(vertex1))
                        poly_verts[index].add(tuple(vertex2))

        # Find the interval of parameter for outer vertex
        for index in range(len(comps)):
            interval1 = RealSet(-infinity, infinity)  # represent t1
            interval2 = RealSet(-infinity, infinity)  # represent t2
            is_doublevar = False
            for i, point in enumerate(comps[index][0]):
                pv = point.variables()
                if len(pv) == 1:
                    temp1 = RealSet(solve(point >= axes[i][0], pv[0])[0][0])
                    temp2 = RealSet(solve(point <= axes[i][1], pv[0])[0][0])
                    temp = temp1.intersection(temp2)
                    if pv[0] == vars[0]:
                        interval1 = interval1.intersection(temp)
                    else:
                        interval2 = interval2.intersection(temp)
                elif len(pv) == 2:
                    sol1 = solve(point >= axes[i][0], pv)
                    sol2 = solve(point <= axes[i][1], pv)
                    is_doublevar = True

            # Find the edge vertices (those that touch the axes)
            find_edge_vertices(0)  # t1 fixed
            find_edge_vertices(1)  # t2 fixed
        return poly_verts

    def plot(self, color='random'):
        """
        Return the plot of ``self`` by constructing a polyhedron from
        vertices in ``self.polygon_vertices()``.

        INPUT:

        - ``color`` -- string or tuple that represent a color (default:
          ``random``); ``random`` means each polygon will be assigned
          a different color. If instead a specific ``color`` is provided,
          then all polygon will be given the same color.

        OUTPUT: Graphics3d Object

        EXAMPLES:

        A tropical surface that consist of only one cell::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y,z> = PolynomialRing(T)
            sage: p1 = x + z
            sage: tv = p1.tropical_variety()
            sage: tv.plot()                                                             # needs sage.plot
            Graphics3d Object

        .. PLOT::
            :width: 300 px

            T = TropicalSemiring(QQ)
            R = PolynomialRing(T, ('x,y,z'))
            x, y, z = R.gen(), R.gen(1), R.gen(2)
            p1 = x + z
            sphinx_plot(p1.tropical_variety().plot())

        A tropical surface with multiple cells that exhibit complex and
        intriguing geometric structures::

            sage: p2 = x^2 + x + y + z + R(1)
            sage: tv = p2.tropical_variety()
            sage: tv.plot()  # long time                                                # needs sage.plot
            Graphics3d Object

        .. PLOT::
            :width: 300 px

            T = TropicalSemiring(QQ)
            R = PolynomialRing(T, ('x,y,z'))
            x, y, z = R.gen(), R.gen(1), R.gen(2)
            p2 = x**2 + x + y + z + R(1)
            sphinx_plot(p2.tropical_variety().plot())
        """
        from random import random
        from sage.plot.graphics import Graphics
        from sage.geometry.polyhedron.constructor import Polyhedron

        if color == 'random':
            colors = []
            for _ in range(self.number_of_components()):
                color = (random(), random(), random())
                colors.append(color)
        elif isinstance(color, str):
            colors = [color] * self.number_of_components()
        else:
            colors = color

        combined_plot = Graphics()
        for i, vertex in self._polygon_vertices().items():
            points = list(vertex)
            plot = Polyhedron(vertices=points).plot(color=colors[i])
            combined_plot += plot
        return combined_plot

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y,z> = PolynomialRing(T)
            sage: (x^4+z^2).tropical_variety()
            Tropical surface of 0*x^4 + 0*z^2
        """
        return f"Tropical surface of {self._poly}"


class TropicalCurve(TropicalVariety):
    r"""
    A tropical curve in `\RR^2`.

    The tropical curve consists of line segments and half-lines, which we
    call edges. These edges are connected in such a way that they form a
    piecewise linear graph embedded in the plane. These edges meet at
    a vertices, where the balancing condition is satisfied. This balancing
    condition ensures that the sum of the outgoing slopes at each vertex
    is zero, reflecting the equilibrium.

    EXAMPLES:

    We define some tropical curves::

        sage: T = TropicalSemiring(QQ, use_min=False)
        sage: R.<x,y> = PolynomialRing(T)
        sage: p1 = x + y + R(0)
        sage: tv1 = p1.tropical_variety(); tv1
        Tropical curve of 0*x + 0*y + 0
        sage: tv1.components()
        [[(t1, t1), [t1 >= 0], 1], [(0, t1), [t1 <= 0], 1], [(t1, 0), [t1 <= 0], 1]]
        sage: tv1.plot()
        Graphics object consisting of 3 graphics primitives

    .. PLOT::
        :width: 300 px

        T = TropicalSemiring(QQ, use_min=False)
        R = PolynomialRing(T, ('x,y'))
        x, y = R.gen(), R.gen(1)
        p1 = x + y + R(0)
        sphinx_plot(p1.tropical_variety().plot())

    ::

        sage: p2 = R(-2)*x^2 + R(-1)*x + R(1/2)*y + R(1/6)
        sage: tv2 = p2.tropical_variety()
        sage: tv2.components()
        [[(1/2*t1 + 5/4, t1), [(-1/3) <= t1], 1],
         [(13/12, t1), [t1 <= (-1/3)], 2],
         [(t1, -1/3), [t1 <= (13/12)], 1]]
        sage: tv2.plot()
        Graphics object consisting of 4 graphics primitives

    .. PLOT::
        :width: 300 px

        T = TropicalSemiring(QQ, use_min=False)
        R = PolynomialRing(T, ('x,y'))
        x, y = R.gen(), R.gen(1)
        p2 = R(-2)*x**2 + R(-1)*x + R(1/2)*y + R(1/6)
        sphinx_plot(p2.tropical_variety().plot())

    When two tropical polynomials are multiplied, the tropical curve of
    the resulting polynomial is the union of the tropical curves of the
    original polynomials::

        sage: p3 = p1 * p2; p3
        (-2)*x^3 + (-2)*x^2*y + (-1)*x^2 + 1/2*x*y + 1/2*y^2 + 1/6*x + 1/2*y + 1/6
        sage: tv3 = p3.tropical_variety()
        sage: tv3.plot()
        Graphics object consisting of 11 graphics primitives

    .. PLOT::
        :width: 300 px

        T = TropicalSemiring(QQ, use_min=False)
        R = PolynomialRing(T, ('x,y'))
        x, y = R.gen(), R.gen(1)
        p1 = x + y + R(0)
        p2 = R(-2)*x**2 + R(-1)*x + R(1/2)*y + R(1/6)
        p3 = p1 * p2
        sphinx_plot(p3.tropical_variety().plot())
    """
    def _axes(self):
        """
        Set the default axes for ``self``.

        This default axes is used for plot of tropical curve and also the
        3d plot of tropical polynomial function. The axes is chosen by first
        find all vertices of this tropical curve. Then we choose the minimum
        and maximum of all x-component in this vertices to be the x-axis.
        The same apply to the y-axis.

        OUTPUT:

        A list of two lists, where the first inner list represent value of
        x-axis and the second inner list represent value of y-axis.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = x^2
            sage: p1.tropical_variety()._axes()
            [[-1, 1], [-1, 1]]
            sage: p2 = R(12)*x*y + R(-2)*y^2 + R(16)*y + R(25)
            sage: p2.tropical_variety()._axes()
            [[-3/2, 1/2], [25/2, 29/2]]
        """
        if self.number_of_components() == 0:
            return [[-1, 1], [-1, 1]]
        if self.number_of_components() <= 2:
            bound = 1
            for comps in self._hypersurface:
                eqns = comps[0]
                temp_operands = []
                for eq in eqns:
                    if not eq.operator():
                        temp_operands.append(eq)
                    else:
                        temp_operands += eq.operands()
                for op in temp_operands:
                    if op.is_numeric():
                        bound = max(abs(op), bound)
            return [[-bound, bound]] * 2

        verts = self.vertices()
        xmin = xmax = list(verts)[0][0]
        for vertex in verts:
            if vertex[0] < xmin:
                xmin = vertex[0]
            elif vertex[0] > xmax:
                xmax = vertex[0]
        ymin = ymax = list(verts)[0][1]
        for vertex in verts:
            if vertex[1] < ymin:
                ymin = vertex[1]
            elif vertex[1] > ymax:
                ymax = vertex[1]
        return [[xmin-1, xmax+1], [ymin-1, ymax+1]]

    def vertices(self):
        r"""
        Return all vertices of ``self``, which is the point where three or
        more edges intersect.

        OUTPUT: a set of `(x,y)` points

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = x + y
            sage: p1.tropical_variety().vertices()
            set()
            sage: p2 = R(-2)*x^2 + R(-1)*x + R(1/2)*y + R(1/6)
            sage: p2.tropical_variety().vertices()
            {(1, -1/2), (7/6, -1/3)}
        """
        if len(self._hypersurface) < 3:
            return set()
        return set(self._vertices_components().keys())

    def _vertices_components(self):
        """
        Return the index of components adjacent to each vertex of ``self``.

        OUTPUT:

        A dictionary where the keys represent the vertices, and the
        values are lists of tuples. Each tuple consists of the index
        of an adjacent edge (component) `e_i` and a number indicating
        the directionality of `e_i` relative to the vertex. The number
        is either -1 or 1.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = R(0) + x + y + x*y + x^2*y + x*y^2
            sage: p1.tropical_variety()._vertices_components()
            {(0, 0): [(0, 1), (1, 1), (2, 1), (3, -1), (4, -1)]}
            sage: p2 = R(2)*x^2 + x*y + R(2)*y^2 + x + R(-1)*y + R(3)
            sage: p2.tropical_variety()._vertices_components()
            {(-2, 0): [(0, -1), (1, 1), (3, 1)],
             (-1, -3): [(2, -1), (4, 1), (5, 1)],
             (-1, 0): [(3, -1), (4, -1), (6, 1)],
             (3, 4): [(6, -1), (7, 1), (8, 1)]}
        """
        comp_vert = {}
        if len(self._hypersurface) >= 3:
            for i, component in enumerate(self._hypersurface):
                parametric_function = component[0]
                v = component[1][0].variables()[0]
                interval = self._parameter_intervals()[i]
                lower = interval[0].lower()
                upper = interval[0].upper()
                if lower != -infinity:
                    x = parametric_function[0].subs(**{str(v): lower})
                    y = parametric_function[1].subs(**{str(v): lower})
                    if (x,y) not in comp_vert:
                        comp_vert[(x,y)] = [(i, 1)]
                    else:
                        comp_vert[(x,y)].append((i, 1))
                if upper != infinity:
                    x = parametric_function[0].subs(**{str(v): upper})
                    y = parametric_function[1].subs(**{str(v): upper})
                    if (x,y) not in comp_vert:
                        comp_vert[(x,y)] = [(i, -1)]
                    else:
                        comp_vert[(x,y)].append((i, -1))
        return comp_vert

    def weight_vectors(self):
        r"""
        Return the weight vectors for all vertices of ``self``.

        Weight vectors are a list of vectors associated with each vertex
        of the curve. Each vector corresponds to an edge emanating from
        that vertex and points in the direction of the edge.

        Suppose `v` is a vertex adjacent to the edges `e_1, \ldots, e_k`
        with respective weights `w_1, \ldots, w_k`. Every edge `e_i` is
        contained in a line (component) defined by an equation. Therefore,
        there exists a unique integer vector `v_i = (\alpha, \beta)` in
        the direction of `e_i` such that `\gcd(\alpha, \beta)=1`. Then,
        each vertex `v` yield the vectors `w_1 v_1, \ldots, w_k v_k`.
        These vectors will satisfy the following balancing condition:
        `\sum_{i=1}^k w_i v_i = 0`.

        OUTPUT:

        A dictionary where the keys represent the vertices, and the values
        are lists of vectors.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = R(-2)*x^2 + R(-1)*x + R(1/2)*y + R(1/6)
            sage: p1.tropical_variety().weight_vectors()
            {(1, -1/2): [(0, 1), (-1, -2), (1, 1)],
             (7/6, -1/3): [(-1, -1), (0, 1), (1, 0)]}

            sage: p2 = R(2)*x^2 + x*y + R(2)*y^2 + x + R(-1)*y + R(3)
            sage: p2.tropical_variety().weight_vectors()
            {(-2, 0): [(-1, -1), (0, 1), (1, 0)],
             (-1, -3): [(-1, -1), (0, 1), (1, 0)],
             (-1, 0): [(-1, 0), (0, -1), (1, 1)],
             (3, 4): [(-1, -1), (0, 1), (1, 0)]}
        """
        from sage.calculus.functional import diff
        from sage.modules.free_module_element import vector
        from sage.arith.misc import gcd

        if not self._vertices_components():
            return {}

        # Calculate the base vector in the direction of each edge
        temp_vectors = []
        par = self._hypersurface[0][1][0].variables()[0]
        for comp in self._hypersurface:
            dx = diff(comp[0][0], par)
            dy = diff(comp[0][1], par)
            multiplier = gcd(QQ(dx), QQ(dy))
            temp_vectors.append(vector([dx/multiplier, dy/multiplier]))

        # Calculate the weight vectors of each vertex
        cov = self._vertices_components()
        result = {}
        for vertex in cov:
            vectors = []
            for comp in cov[vertex]:
                weight = self._hypersurface[comp[0]][2]
                vectors.append(weight*comp[1]*temp_vectors[comp[0]])
            result[vertex] = vectors
        return result

    def is_smooth(self):
        r"""
        Return ``True`` if ``self`` is smooth and ``False`` otherwise.

        Suppose `C` is a tropical curve of degree `d`. A tropical curve
        `C` is smooth if the dual subdivision of `C` consists of `d^2`
        triangles each having unit area `1/2`. This is equivalent with
        `C` having `d^2` vertices. These vertices are necessarily
        trivalent (has three adjacent edges).

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = x^2 + x + R(1)
            sage: p1.tropical_variety().is_smooth()
            False
            sage: p2 = R(2)*x^2 + x*y + R(2)*y^2 + x + R(-1)*y + R(3)
            sage: p2.tropical_variety().is_smooth()
            True
        """
        return len(self.vertices()) == self._poly.degree() ** 2

    def is_simple(self):
        r"""
        Return ``True`` if ``self`` is simple and ``False`` otherwise.

        A tropical curve `C` is called simple if each vertex is either
        trivalent or is locally the intersection of two line segments.
        Equivalently, `C` is simple if the corresponding subdivision
        consists only of triangles and parallelograms.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = R(0) + x + y + x*y + x^2*y + x*y^2
            sage: p1.tropical_variety().is_simple()
            False
            sage: p2 = R(2)*x^2 + x*y + R(2)*y^2 + x + R(-1)*y + R(3)
            sage: p2.tropical_variety().is_simple()
            True
        """
        vov = self.weight_vectors()
        for vertex in self.vertices():
            if len(vov[vertex]) > 4:
                return False
            if len(vov[vertex]) == 4:
                if any(-v not in vov[vertex] for v in vov[vertex]):
                    return False
        return True

    def genus(self):
        r"""
        Return the genus of ``self``.

        Let `t(C)` be the number of trivalent vertices, and let `r(C)` be
        the number of unbounded edges of `C`. The genus of simple tropical
        curve `C` is defined by the formula:

        .. MATH::

            g(C) = \frac{1}{2}t(C) - \frac{1}{2}r(C) + 1.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = x^2 + y^2 + x*y
            sage: p1.tropical_variety().genus()
            1
            sage: p2 = R(2)*x^2 + x*y + R(2)*y^2 + x + R(-1)*y + R(3)
            sage: p2.tropical_variety().genus()
            0

        TESTS::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = R(0) + y + x^2*y + x*y^2
            sage: p1.tropical_variety().genus()
            Traceback (most recent call last):
            ...
            ValueError: Tropical curve of 0*x^2*y + 0*x*y^2 + 0*y + 0 is not simple
        """
        if not self.is_simple():
            raise ValueError(f"{self} is not simple")
        trivalent = 0  # number of trivalent vertices
        for vectors in self.weight_vectors().values():
            if len(vectors) == 3:
                trivalent += 1
        unbounded = 0  # number of unbounded edges
        for component in self._hypersurface:
            if len(component[1]) == 1:
                unbounded += 1
        return trivalent//2 - unbounded//2 + 1

    def contribution(self):
        r"""
        Return the contribution of ``self``.

        The contribution of a simple curve `C` is defined as the product
        of the normalized areas of all triangles in the corresponding
        dual subdivision. We just multiply positive integers attached to
        the trivalent vertices. The contribution of a trivalent vertex
        equals `w_1w_2|\det(v_1,v_2)|`, with `w_i` are the weights of
        the adjacent edges and `v_i` are their weight vectors. That
        formula is independent of the choice made because of the
        balancing condition `w_1v_1+w_2v_2+w_3v_3=0`.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = R(2)*x^2 + x*y + R(2)*y^2 + x + R(-1)*y + R(3)
            sage: p1.tropical_variety().contribution()
            1
            sage: p2 = R(-1/3)*x^2 + R(1)*x*y + R(1)*y^2 + R(-1/3)*x + R(1/3)
            sage: p2.tropical_variety().contribution()
            16

        TESTS::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = R(0) + x + x^2*y + x*y^2
            sage: p1.tropical_variety().contribution()
            Traceback (most recent call last):
            ...
            ValueError: Tropical curve of 0*x^2*y + 0*x*y^2 + 0*x + 0 is not simple
        """
        if not self.is_simple():
            raise ValueError(f"{self} is not simple")
        result = 1
        voc = self._vertices_components()
        vov = self.weight_vectors()
        for vertex in vov:
            if len(vov[vertex]) == 3:
                u1 = vov[vertex][0]
                u2 = vov[vertex][1]
                index1 = voc[vertex][0][0]
                index2 = voc[vertex][1][0]
                w1 = self._hypersurface[index1][2]
                w2 = self._hypersurface[index2][2]
                det = u1[0]*u2[1] - u1[1]*u2[0]
                result *= w1 * w2 * abs(det)
        return result

    def _parameter_intervals(self):
        r"""
        Return the intervals of each component's parameter of ``self``.

        OUTPUT: a list of ``RealSet``

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: p1 = y + y^2
            sage: p1.tropical_variety()._parameter_intervals()
            [(-oo, +oo)]
            sage: p2 = x^2 + R(-1)*x*y + R(-1)*x + R(1/3)
            sage: p2.tropical_variety()._parameter_intervals()
            [(-oo, 0], [0, +oo), [-1, 4/3], (-oo, 0], [0, +oo)]
        """
        from sage.sets.real_set import RealSet

        intervals = []
        R = self._poly.parent().base().base_ring()
        for component in self._hypersurface:
            if len(component[1]) == 1:
                interval = RealSet(component[1][0])
            else:
                lower = component[1][0].left()
                upper = component[1][1].right()
                if lower == -infinity:
                    interval = RealSet(-infinity, infinity)
                else:
                    interval = RealSet([R(lower), R(upper)])
            intervals.append(interval)
        return intervals

    def plot(self):
        """
        Return the plot of ``self``.

        Generates a visual representation of the tropical curve in cartesian
        coordinates. The plot shows piecewise-linear segments representing
        each components. The axes are centered around the vertices.

        OUTPUT:

        A Graphics object. The weight of the component will be written if it
        is greater or equal than 2. The weight is written near the vertex.

        EXAMPLES:

        A polynomial with only two terms will give one straight line::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: (y+R(1)).tropical_variety().components()
            [[(t1, 1), [-Infinity < t1, t1 < +Infinity], 1]]
            sage: (y+R(1)).tropical_variety().plot()                                    # needs sage.plot
            Graphics object consisting of 1 graphics primitive

        .. PLOT::
            :width: 300 px

            T = TropicalSemiring(QQ)
            R = PolynomialRing(T, ('x,y'))
            x, y = R.gen(), R.gen(1)
            sphinx_plot((y+R(1)).tropical_variety().plot())

        An intriguing and fascinating tropical curve can be obtained with
        a more complex tropical polynomial::

            sage: p1 = R(1) + R(2)*x + R(3)*y + R(6)*x*y + R(10)*x*y^2
            sage: p1.tropical_variety().components()
            [[(-1, t1), [-2 <= t1], 1],
            [(t1, -2), [-1 <= t1], 1],
            [(t1 + 1, t1), [-4 <= t1, t1 <= -2], 1],
            [(t1, -4), [t1 <= -3], 2],
            [(-t1 - 7, t1), [t1 <= -4], 1]]
            sage: p1.tropical_variety().plot()                                          # needs sage.plot
            Graphics object consisting of 6 graphics primitives

        .. PLOT::
            :width: 300 px

            T = TropicalSemiring(QQ)
            R = PolynomialRing(T, ('x,y'))
            x, y = R.gen(), R.gen(1)
            p1 = R(1) + R(2)*x + R(3)*y + R(6)*x*y + R(10)*x*y**2
            sphinx_plot(p1.tropical_variety().plot())

        Another tropical polynomial with numerous components, resulting
        in a more intricate structure::

            sage: p2 = (x^6 + R(4)*x^4*y^2 + R(2)*x^3*y^3 + R(3)*x^2*y^4
            ....:       + x*y^5 + R(7)*x^2 + R(5)*x*y + R(3)*y^2 + R(2)*x
            ....:       + y + R(10))
            sage: p2.tropical_variety().plot()  # long time                             # needs sage.plot
            Graphics object consisting of 11 graphics primitives

        .. PLOT::
            :width: 300 px

            T = TropicalSemiring(QQ)
            R = PolynomialRing(T, ('x,y'))
            x, y = R.gen(), R.gen(1)
            p2 = (x**6 + R(4)*x**4*y**2 + R(2)*x**3*y**3 + R(3)*x**2*y**4
                  + x*y**5 + R(7)*x**2 + R(5)*x*y + R(3)*y**2 + R(2)*x
                  + y + R(10))
            sphinx_plot(p2.tropical_variety().plot())

        ::

            sage: p3 = (R(8) + R(4)*x + R(2)*y + R(1)*x^2 + x*y + R(1)*y^2
            ....:       + R(2)*x^3 + x^2*y + x*y^2 + R(4)*y^3 + R(8)*x^4
            ....:       + R(4)*x^3*y + x^2*y^2 + R(2)*x*y^3 + y^4)
            sage: p3.tropical_variety().plot() # long time
            Graphics object consisting of 23 graphics primitives

        .. PLOT::
            :width: 300 px

            T = TropicalSemiring(QQ)
            R = PolynomialRing(T, ('x,y'))
            x, y = R.gen(), R.gen(1)
            p3 = (R(8) + R(4)*x + R(2)*y + R(1)*x**2 + x*y + R(1)*y**2
                 + R(2)*x**3 + x**2*y + x*y**2 + R(4)*y**3 + R(8)*x**4
                 + R(4)*x**3*y + x**2*y**2 + R(2)*x*y**3 + y**4)
            sphinx_plot(p3.tropical_variety().plot())
        """
        from sage.plot.plot import plot
        from sage.plot.text import text
        from sage.plot.graphics import Graphics
        from sage.plot.plot import parametric_plot

        if not self._hypersurface:
            return plot(lambda x: float('nan'), {-1, 1})

        combined_plot = Graphics()
        large_int = 100
        intervals = self._parameter_intervals()
        for i, component in enumerate(self._hypersurface):
            var = component[1][0].variables()[0]
            parametric_function = component[0]
            order = component[2]
            interval = intervals[i]
            if interval[0].lower() == -infinity:
                lower = interval[0].upper() - large_int
                upper = interval[0].upper()
                midpoint = upper - 0.5
            elif interval[0].upper() == infinity:
                lower = interval[0].lower()
                upper = interval[0].lower() + large_int
                midpoint = lower + 0.5
            else:
                lower = interval[0].lower()
                upper = interval[0].upper()
                midpoint = (lower+upper) / 2

            if (lower == infinity) and (upper == infinity):
                midpoint = 0
                plot = parametric_plot(parametric_function, (var, -large_int,
                                       large_int), color='red')
            else:
                plot = parametric_plot(parametric_function, (var, lower, upper),
                                       color='red')

            # Add the order if it is greater than or equal to 2
            if component[2] > 1:
                point = []
                for eq in component[0]:
                    value = eq.subs(**{str(var): midpoint})
                    point.append(value)
                text_order = text(str(order), (point[0], point[1]),
                                  fontsize=16, color='black')
                combined_plot += plot + text_order
            else:
                combined_plot += plot

        # Set default axes
        axes = self._axes()
        xmin, xmax = axes[0][0], axes[0][1]
        ymin, ymax = axes[1][0], axes[1][1]
        combined_plot.set_axes_range(xmin=xmin, xmax=xmax,
                                     ymin=ymin, ymax=ymax)
        return combined_plot

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: T = TropicalSemiring(QQ)
            sage: R.<x,y> = PolynomialRing(T)
            sage: (x^2+R(0)).tropical_variety()
            Tropical curve of 0*x^2 + 0
        """
        return f"Tropical curve of {self._poly}"



# sage_setup: distribution = sagemath-categories
"""
Power Series Methods

The class ``PowerSeries_poly`` provides additional methods for univariate power series.
"""
from sage.rings.power_series_ring_element cimport PowerSeries
from sage.structure.element cimport Element
from sage.rings.infinity import infinity

try:
    from sage.libs.pari.all import pari_gen, PariError
except ImportError:
    pari_gen = ()
    PariError = ()


cdef class PowerSeries_poly(PowerSeries):

    def __init__(self, parent, f=0, prec=infinity, int check=1, is_gen=0):
        """
        EXAMPLES::

            sage: R.<q> = PowerSeriesRing(CC); R                                        # needs sage.rings.real_mpfr
            Power Series Ring in q over Complex Field with 53 bits of precision
            sage: loads(q.dumps()) == q                                                 # needs sage.rings.real_mpfr
            True

            sage: R.<t> = QQ[[]]
            sage: f = 3 - t^3 + O(t^5)
            sage: a = f^3; a
            27 - 27*t^3 + O(t^5)
            sage: b = f^-3; b
            1/27 + 1/27*t^3 + O(t^5)
            sage: a*b
            1 + O(t^5)

        Check that :issue:`22216` is fixed::

            sage: R.<T> = PowerSeriesRing(QQ)
            sage: R(pari('1 + O(T)'))                                                   # needs sage.libs.pari
            1 + O(T)
            sage: R(pari('1/T + O(T)'))                                                 # needs sage.libs.pari
            Traceback (most recent call last):
            ...
            ValueError: series has negative valuation
        """
        R = parent._poly_ring()
        if isinstance(f, Element):
            if (<Element>f)._parent is R:
                pass
            elif (<Element>f)._parent == R.base_ring():
                f = R([f])
            elif isinstance(f, PowerSeries):  # not only PowerSeries_poly
                prec = (<PowerSeries>f)._prec
                f = R(f.polynomial())
            else:
                if f:
                    f = R(f, check=check)
                else:
                    f = R(None)
        elif isinstance(f, pari_gen) and f.type() == 't_SER':
            if f._valp() < 0:
                raise ValueError('series has negative valuation')
            if prec is infinity:
                prec = f.length() + f._valp()
            f = R(f.truncate())
        else:
            if f:
                f = R(f, check=check)
            else: # None is supposed to yield zero
                f = R(None)

        self.__f = f
        if check and not (prec is infinity):
            self.__f = self.__f.truncate(prec)
        PowerSeries.__init__(self, parent, prec, is_gen)

    def __hash__(self):
        """
        Return a hash of ``self``.

        EXAMPLES::

            sage: R.<t> = ZZ[[]]
            sage: hash(t) == hash(R.gen())
            True
            sage: hash(t) != hash(R.one())
            True
        """
        return hash(self.__f)

    def __reduce__(self):
        """
        Used for pickling.

        EXAMPLES::

            sage: A.<z> = RR[[]]
            sage: f = z - z^3 + O(z^10)
            sage: f == loads(dumps(f)) # indirect doctest
            True
        """
        return self.__class__, (self._parent, self.__f, self._prec, self._is_gen)

    def polynomial(self):
        """
        Return the underlying polynomial of ``self``.

        EXAMPLES::

            sage: R.<t> = GF(7)[[]]
            sage: f = 3 - t^3 + O(t^5)
            sage: f.polynomial()
            6*t^3 + 3
        """
        return self.__f

    def valuation(self):
        """
        Return the valuation of ``self``.

        EXAMPLES::

            sage: R.<t> = QQ[[]]
            sage: (5 - t^8 + O(t^11)).valuation()
            0
            sage: (-t^8 + O(t^11)).valuation()
            8
            sage: O(t^7).valuation()
            7
            sage: R(0).valuation()
            +Infinity
        """
        if self.__f == 0:
            return self._prec

        return self.__f.valuation()

    def degree(self):
        """
        Return the degree of the underlying polynomial of ``self``.

        That is, if ``self`` is of the form `f(x) + O(x^n)`, we return
        the degree of `f(x)`. Note that if `f(x)` is `0`, we return `-1`,
        just as with polynomials.

        EXAMPLES::

            sage: R.<t> = ZZ[[]]
            sage: (5 + t^3 + O(t^4)).degree()
            3
            sage: (5 + O(t^4)).degree()
            0
            sage: O(t^4).degree()
            -1
        """
        return self.__f.degree()

    def __bool__(self):
        """
        Return ``True`` if ``self`` is nonzero, and ``False`` otherwise.

        EXAMPLES::

            sage: R.<t> = GF(11)[[]]
            sage: bool(1 + t + O(t^18))
            True
            sage: bool(R(0))
            False
            sage: bool(O(t^18))
            False
        """
        return not not self.__f

    def __call__(self, *x, **kwds):
        """
        Evaluate the series at `x=a`.

        INPUT:

        - ``x``:

          - a tuple of elements the first of which can be meaningfully
            substituted in ``self``, with the remainder used for substitution
            in the coefficients of ``self``.

          - a dictionary for kwds:value pairs. If the variable name of
            ``self`` is a keyword it is substituted for.  Other keywords
            are used for substitution in the coefficients of ``self``.

        OUTPUT: the value of ``self`` after substitution

        EXAMPLES::

            sage: R.<t> = ZZ[[]]
            sage: f = t^2 + t^3 + O(t^6)
            sage: f(t^3)
            t^6 + t^9 + O(t^18)
            sage: f(t=t^3)
            t^6 + t^9 + O(t^18)
            sage: f(f)
            t^4 + 2*t^5 + 2*t^6 + 3*t^7 + O(t^8)
            sage: f(f)(f) == f(f(f))
            True

        The following demonstrates that the problems raised in :issue:`3979`
        and :issue:`5367` are solved::

            sage: [f(t^2 + O(t^n)) for n in [9, 10, 11]]
            [t^4 + t^6 + O(t^11), t^4 + t^6 + O(t^12), t^4 + t^6 + O(t^12)]
            sage: f(t^2)
            t^4 + t^6 + O(t^12)

        It is possible to substitute a series for which only the precision
        is defined::

            sage: f(O(t^5))
            O(t^10)

        or to substitute a polynomial (the result belonging to the power
        series ring over the same base ring)::

            sage: P.<z> = ZZ[]
            sage: g = f(z + z^3); g
            z^2 + z^3 + 2*z^4 + 3*z^5 + O(z^6)
            sage: g.parent()
            Power Series Ring in z over Integer Ring

        A series defined over another ring can be substituted::

            sage: S.<u> = GF(7)[[]]
            sage: f(2*u + u^3 + O(u^5))
            4*u^2 + u^3 + 4*u^4 + 5*u^5 + O(u^6)

        As can a `p`-adic integer as long as the coefficient ring is compatible::

            sage: f(100 + O(5^7))                                                       # needs sage.rings.padics
            5^4 + 3*5^5 + 4*5^6 + 2*5^7 + 2*5^8 + O(5^9)
            sage: f.change_ring(Zp(5))(100 + O(5^7))                                    # needs sage.rings.padics
            5^4 + 3*5^5 + 4*5^6 + 2*5^7 + 2*5^8 + O(5^9)
            sage: f.change_ring(Zp(5))(100 + O(2^7))                                    # needs sage.rings.padics
            Traceback (most recent call last):
            ...
            ValueError: Cannot substitute this value

        To substitute a value it must have valuation at least 1::

            sage: f(0)
            0
            sage: f(1 + t)
            Traceback (most recent call last):
            ...
            ValueError: Can only substitute elements of positive valuation
            sage: f(2 + O(5^3))                                                         # needs sage.rings.padics
            Traceback (most recent call last):
            ...
            ValueError: Can only substitute elements of positive valuation
            sage: f(t^-2)
            Traceback (most recent call last):
            ...
            ValueError: Can only substitute elements of positive valuation

        Unless, of course, it is being substituted in a series with infinite
        precision, i.e., a polynomial::

            sage: g = t^2 + t^3
            sage: g(1 + t + O(t^2))
            2 + 5*t + O(t^2)
            sage: g(3)
            36

        Arguments beyond the first can refer to the base ring::

            sage: P.<x> = GF(5)[]
            sage: Q.<y> = P[[]]
            sage: h = (1 - x*y)^-1 + O(y^7); h
            1 + x*y + x^2*y^2 + x^3*y^3 + x^4*y^4 + x^5*y^5 + x^6*y^6 + O(y^7)
            sage: h(y^2, 3)
            1 + 3*y^2 + 4*y^4 + 2*y^6 + y^8 + 3*y^10 + 4*y^12 + O(y^14)

        These secondary values can also be specified using keywords::

            sage: h(y=y^2, x=3)
            1 + 3*y^2 + 4*y^4 + 2*y^6 + y^8 + 3*y^10 + 4*y^12 + O(y^14)
            sage: h(y^2, x=3)
            1 + 3*y^2 + 4*y^4 + 2*y^6 + y^8 + 3*y^10 + 4*y^12 + O(y^14)
        """
        P = self.parent()

        if len(kwds) >= 1:
            name = P.variable_name()
            if name in kwds: # a keyword specifies the power series generator
                if x:
                    raise ValueError("must not specify %s keyword and positional argument" % name)
                a = self(kwds[name])
                del kwds[name]
                try:
                    return a(**kwds)
                except TypeError:
                    return a
            elif x:       # both keywords and positional arguments
                a = self(*x)
                try:
                    return a(**kwds)
                except TypeError:
                    return a
            else:                  # keywords but no positional arguments
                return P(self.__f(**kwds)).add_bigoh(self._prec)

        if len(x) == 0:
            return self

        if isinstance(x[0], tuple):
            x = x[0]
        a = x[0]

        s = self._prec
        if s == infinity:
            return self.__f(x)

        Q = a.parent()

        from sage.rings.padics.padic_generic import pAdicGeneric
        padic = isinstance(Q, pAdicGeneric)
        if padic:
            p = Q.prime()

        try:
            t = a.valuation()
        except (TypeError, AttributeError):
            if a.is_zero():
                t = infinity
            else:
                t = 0

        if t == infinity:
            return self[0]

        if t <= 0:
            raise ValueError("Can only substitute elements of positive valuation")

        if not Q.has_coerce_map_from(P.base_ring()):
            from sage.structure.element import canonical_coercion
            try:
                R = canonical_coercion(P.base_ring()(0), Q.base_ring()(0))[0].parent()
                self = self.change_ring(R)
            except TypeError:
                raise ValueError("Cannot substitute this value")

        r = (self - self[0]).valuation()
        if r == s:                 # self is constant + O(x^s)
            if padic:
                from sage.rings.big_oh import O
                return self[0] + O(p**(s*t))
            else:
                return P(self[0]).add_bigoh(s*t)

        try:
            u = a.prec()
        except AttributeError:
            u = a.precision_absolute()
        n = (s - r + 1)*t
        if n < u:
            a = a.add_bigoh(n)
            x = list(x)
            x[0] = a
            x = tuple(x)
        return self.__f(x)

    def _unsafe_mutate(self, i, value):
        """
        Sage assumes throughout that commutative ring elements are immutable.
        This is relevant for caching, etc.  But sometimes you need to change
        a power series and you really know what you're doing.  That's
        when this function is for you.

        ** DO NOT USE THIS ** unless you know what you're doing.

        EXAMPLES::

            sage: R.<t> = GF(7)[[]]
            sage: f = 3 + 6*t^3 + O(t^5)
            sage: f._unsafe_mutate(0, 5)
            sage: f
            5 + 6*t^3 + O(t^5)
            sage: f._unsafe_mutate(2, 1); f
            5 + t^2 + 6*t^3 + O(t^5)

        - Mutating can even bump up the precision::

            sage: f._unsafe_mutate(6, 1); f
            5 + t^2 + 6*t^3 + t^6 + O(t^7)
            sage: f._unsafe_mutate(0, 0); f
            t^2 + 6*t^3 + t^6 + O(t^7)
            sage: f._unsafe_mutate(1, 0); f
            t^2 + 6*t^3 + t^6 + O(t^7)
            sage: f._unsafe_mutate(11,0); f
            t^2 + 6*t^3 + t^6 + O(t^12)

            sage: g = t + O(t^7)
            sage: g._unsafe_mutate(1,0); g
            O(t^7)
        """
        self.__f._unsafe_mutate(i, value)
        self._prec = max(self._prec, i+1)

    def __getitem__(self, n):
        """
        Return the ``n``-th coefficient of ``self``.

        This returns 0 for negative coefficients and raises an
        :exc:`IndexError` if trying to access beyond known coefficients.

        If ``n`` is a slice object ``[:k]``, this will return a power
        series of the same precision, whose coefficients are the same
        as ``self`` for those indices in the slice, and 0 otherwise.
        Other kinds of slicing are not allowed.

        EXAMPLES::

            sage: R.<t> = QQ[[]]
            sage: f = 3/2 - 17/5*t^3 + O(t^5)
            sage: f[3]
            -17/5
            sage: f[-2]
            0
            sage: f[4]
            0
            sage: f[5]
            Traceback (most recent call last):
            ...
            IndexError: coefficient not known

        Using slices::

            sage: R.<t> = ZZ[[]]
            sage: f = (2-t)^5; f
            32 - 80*t + 80*t^2 - 40*t^3 + 10*t^4 - t^5
            sage: f[:4]
            32 - 80*t + 80*t^2 - 40*t^3
            sage: f = 1 + t^3 - 4*t^4 + O(t^7); f
            1 + t^3 - 4*t^4 + O(t^7)
            sage: f[:4]
            1 + t^3 + O(t^7)

        TESTS::

            sage: f[1:4]
            Traceback (most recent call last):
            ...
            IndexError: polynomial slicing with a start is not defined
        """
        if isinstance(n, slice):
            return PowerSeries_poly(self._parent, self.polynomial()[n],
                                    prec=self._prec, check=False)
        elif n < 0:
            return self.base_ring().zero()
        elif n > self.__f.degree():
            if self._prec > n:
                return self.base_ring().zero()
            else:
                raise IndexError("coefficient not known")
        return self.__f[n]

    def __iter__(self):
        """
        Return an iterator over the coefficients of this power series.

        EXAMPLES::

            sage: R.<t> = QQ[[]]
            sage: f = t + 17/5*t^3 + 2*t^4 + O(t^5)
            sage: [a for a in f]
            [0, 1, 0, 17/5, 2]
        """
        return iter(self.__f)

    def __neg__(self):
        """
        Return the negative of this power series.

        EXAMPLES::

            sage: R.<t> = QQ[[]]
            sage: f = t + 17/5*t^3 + 2*t^4 + O(t^5)
            sage: -f
            -t - 17/5*t^3 - 2*t^4 + O(t^5)
        """
        return PowerSeries_poly(self._parent, -self.__f,
                                         self._prec, check=False)

    cpdef _add_(self, right_m):
        """
        EXAMPLES::

            sage: R.<x> = PowerSeriesRing(ZZ)
            sage: f = x^4 + O(x^5); f
            x^4 + O(x^5)
            sage: g = x^2 + O(x^3); g
            x^2 + O(x^3)
            sage: f+g
            x^2 + O(x^3)

        TESTS:

        In the past this could die with EXC_BAD_ACCESS (:issue:`8029`)::

            sage: # needs sage.rings.real_mpfr
            sage: A.<x> = RR['x']
            sage: B.<t> = PowerSeriesRing(A)
            sage: 1. + O(t)
            1.00000000000000 + O(t)
            sage: 1. + O(t^2)
            1.00000000000000 + O(t^2)
            sage: 1. + O(t^3)
            1.00000000000000 + O(t^3)
            sage: 1. + O(t^4)
            1.00000000000000 + O(t^4)
        """
        cdef PowerSeries_poly right = <PowerSeries_poly>right_m
        return PowerSeries_poly(self._parent, self.__f + right.__f,
                                self.common_prec_c(right), check=True)

    cpdef _sub_(self, right_m):
        """
        Return the difference of two power series.

        EXAMPLES::

            sage: k.<w> = ZZ[]
            sage: R.<t> = k[[]]
            sage: w*t^2 -w*t +13 - (w*t^2 + w*t)
            13 - 2*w*t
        """
        cdef PowerSeries_poly right = <PowerSeries_poly>right_m
        return PowerSeries_poly(self._parent, self.__f - right.__f,
                                self.common_prec_c(right), check=True)

    cpdef _mul_(self, right_r):
        """
        Return the product of two power series.

        EXAMPLES::

            sage: k.<w> = ZZ[[]]
            sage: (1+17*w+15*w^3+O(w^5))*(19*w^10+O(w^12))
            19*w^10 + 323*w^11 + O(w^12)
        """
        prec = self._mul_prec(right_r)
        return PowerSeries_poly(self._parent,
                                self.__f * (<PowerSeries_poly>right_r).__f,
                                prec=prec,
                                check=True)  # check, since truncation may be needed

    cpdef _rmul_(self, Element c):
        """
        Multiply ``self`` on the right by a scalar.

        EXAMPLES::

            sage: R.<t> = GF(7)[[]]
            sage: f = t + 3*t^4 + O(t^11)
            sage: f * GF(7)(3)
            3*t + 2*t^4 + O(t^11)
        """
        return PowerSeries_poly(self._parent, self.__f * c, self._prec, check=False)

    cpdef _lmul_(self, Element c):
        """
        Multiply ``self`` on the left by a scalar.

        EXAMPLES::

            sage: R.<t> = GF(11)[[]]
            sage: f = 1 + 3*t^4 + O(t^120)
            sage: 2 * f
            2 + 6*t^4 + O(t^120)
        """
        return PowerSeries_poly(self._parent, c * self.__f, self._prec, check=False)

    def __lshift__(PowerSeries_poly self, n):
        """
        Shift ``self`` to the left by ``n``, i.e. multiply by `x^n`.

        EXAMPLES::

            sage: R.<t> = QQ[[]]
            sage: f = 1 + t + t^4
            sage: f << 1
            t + t^2 + t^5
        """
        if n:
            return PowerSeries_poly(self._parent, self.__f << n, self._prec + n)
        else:
            return self

    def __rshift__(PowerSeries_poly self, n):
        """
        Shift ``self`` to the right by ``n``, i.e. multiply by `x^{-n}` and
        remove any terms of negative exponent.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: R.<t> = GF(2)[[]]
            sage: f = t + t^4 + O(t^7)
            sage: f >> 1
            1 + t^3 + O(t^6)
            sage: f >> 10
            O(t^0)
        """
        if n:
            return PowerSeries_poly(self._parent, self.__f >> n, max(0,self._prec - n))
        else:
            return self

    def __invert__(self):
        """
        Return the inverse of the power series (i.e., a series `Y` such
        that `XY = 1`).

        The first nonzero coefficient must be a unit in
        the coefficient ring. If the valuation of the series is positive or
        `X` is not a unit, this function will return a
        :class:`sage.rings.laurent_series_ring_element.LaurentSeries`.

        EXAMPLES::

            sage: R.<q> = QQ[[]]
            sage: 1/(1+q + O(q**2))
            1 - q + O(q^2)
            sage: 1/(1+q)
            1 - q + q^2 - q^3 + q^4 - q^5 + q^6 - q^7 + q^8 - q^9 + q^10 - q^11 + q^12 - q^13 + q^14 - q^15 + q^16 - q^17 + q^18 - q^19 + O(q^20)
            sage: prec = R.default_prec(); prec
            20
            sage: 1/(1+q) + O(q^5)
            1 - q + q^2 - q^3 + q^4 + O(q^5)

        ::

            sage: 1/(q + q^2) + O(q^4)
            q^-1 - 1 + q - q^2 + q^3 + O(q^4)
            sage: g = 1/(q + q^2 + O(q^5))
            sage: g; g.parent()
            q^-1 - 1 + q - q^2 + O(q^3)
            Laurent Series Ring in q over Rational Field

        ::

            sage: 1/g
            q + q^2 + O(q^5)
            sage: (1/g).parent()
            Laurent Series Ring in q over Rational Field

        ::

            sage: 1/(2 + q) + O(q^5)
            1/2 - 1/4*q + 1/8*q^2 - 1/16*q^3 + 1/32*q^4 + O(q^5)

        ::

            sage: R.<q> = PowerSeriesRing(QQ, name='q', default_prec=5)
            sage: f = 1 + q + q^2 + O(q^50)
            sage: f/10
            1/10 + 1/10*q + 1/10*q^2 + O(q^50)
            sage: f/(10+q)
            1/10 + 9/100*q + 91/1000*q^2 - 91/10000*q^3 + 91/100000*q^4 + O(q^5)

        ::

            sage: R.<t> = PowerSeriesRing(QQ, sparse=True)
            sage: u = 17 + 3*t^2 + 19*t^10 + O(t^12)
            sage: v = ~u; v
            1/17 - 3/289*t^2 + 9/4913*t^4 - 27/83521*t^6 + 81/1419857*t^8 - 1587142/24137569*t^10 + O(t^12)
            sage: u*v
            1 + O(t^12)

        If we try a nonzero, non-unit constant term, we end up in
        the fraction field, i.e. the Laurent series ring::

            sage: R.<t> = PowerSeriesRing(ZZ)
            sage: ~R(2)
            1/2
            sage: parent(~R(2))
            Laurent Series Ring in t over Rational Field

        As for units, we stay in the power series ring::

            sage: ~R(-1)
            -1
            sage: parent(~R(-1))
            Power Series Ring in t over Integer Ring

        However, inversion of non-unit elements must fail when the underlying
        ring is not an integral domain::

            sage: R = IntegerModRing(8)
            sage: P.<s> = R[[]]
            sage: ~P(2)
            Traceback (most recent call last):
            ...
            ValueError: must be an integral domain
        """
        if self.is_one():
            return self
        prec = self.prec()
        if prec is infinity:
            if self.degree() > 0:
                prec = self._parent.default_prec()
            else:
                # constant series
                a = self[0]
                if not a.is_unit():
                    from sage.categories.integral_domains import IntegralDomains
                    if self._parent in IntegralDomains():
                        R = self._parent.fraction_field()
                        return 1 / R(a)
                    else:
                        raise ValueError('must be an integral domain')
                try:
                    a = a.inverse_unit()
                except (AttributeError, NotImplementedError):
                    a = self._parent.base_ring()(~a)
                return self._parent(a, prec=infinity)

        if self.valuation() > 0:
            u = ~self.valuation_zero_part()    # inverse of unit part
            R = self._parent.laurent_series_ring()
            return R(u, -self.valuation())

        return self._parent(self.truncate().inverse_series_trunc(prec), prec=prec)

    def truncate(self, prec=infinity):
        """
        The polynomial obtained from power series by truncation at
        precision ``prec``.

        EXAMPLES::

            sage: R.<I> = GF(2)[[]]
            sage: f = 1/(1+I+O(I^8)); f
            1 + I + I^2 + I^3 + I^4 + I^5 + I^6 + I^7 + O(I^8)
            sage: f.truncate(5)
            I^4 + I^3 + I^2 + I + 1
        """
        if prec is infinity:
            return self.__f
        else:
            return self.__f.truncate(prec)

    cdef _inplace_truncate(self, long prec):
        """
        Truncate ``self`` to precision ``prec`` in place.

        .. NOTE::

            This is very unsafe, since power series are supposed to
            be immutable in Sage. Use at your own risk!
        """
        self.__f = self.__f._inplace_truncate(prec)
        self.prec = prec
        return self

    def truncate_powerseries(self, long prec):
        r"""
        Given input ``prec`` = `n`, returns the power series of degree
        `< n` which is equivalent to ``self`` modulo `x^n`.

        EXAMPLES::

            sage: R.<I> = GF(2)[[]]
            sage: f = 1/(1+I+O(I^8)); f
            1 + I + I^2 + I^3 + I^4 + I^5 + I^6 + I^7 + O(I^8)
            sage: f.truncate_powerseries(5)
            1 + I + I^2 + I^3 + I^4 + O(I^5)
        """
        return PowerSeries_poly(self._parent, self.__f.truncate(prec),
                                min(self._prec, prec), check=False)

    def list(self):
        """
        Return the list of known coefficients for ``self``.

        This is just the list of coefficients of the underlying
        polynomial, so in particular, need not have length equal to
        ``self.prec()``.

        EXAMPLES::

            sage: R.<t> = ZZ[[]]
            sage: f = 1 - 5*t^3 + t^5 + O(t^7)
            sage: f.list()
            [1, 0, 0, -5, 0, 1]
        """
        return self.__f.list()

    def monomial_coefficients(self):
        """
        Return a dictionary of coefficients for ``self``.

        This is simply a dict for the underlying polynomial, so need
        not have keys corresponding to every number smaller than
        ``self.prec()``.

        EXAMPLES::

            sage: R.<t> = ZZ[[]]
            sage: f = 1 + t^10 + O(t^12)
            sage: f.monomial_coefficients()
            {0: 1, 10: 1}

        ``dict`` is an alias::

            sage: f.dict()
            {0: 1, 10: 1}
        """
        return self.__f.monomial_coefficients()

    dict = monomial_coefficients

    def _derivative(self, var=None):
        """
        Return the derivative of this power series with respect
        to the variable ``var``.

        If ``var`` is ``None`` or is the generator of this ring, we
        take the derivative with respect to the generator.

        Otherwise, we call ``_derivative(var)`` on each coefficient of
        the series.

        .. SEEALSO::

            ``self.derivative()``

        EXAMPLES::

            sage: R.<t> = PowerSeriesRing(QQ, sparse=True)
            sage: f = 2 + 3*t^2 + t^100000 + O(t^10000000); f
            2 + 3*t^2 + t^100000 + O(t^10000000)
            sage: f._derivative()
            6*t + 100000*t^99999 + O(t^9999999)
            sage: f._derivative(t)
            6*t + 100000*t^99999 + O(t^9999999)

            sage: R.<x> = PolynomialRing(ZZ)
            sage: S.<y> = PowerSeriesRing(R, sparse=True)
            sage: f = x^3*y^4 + O(y^5)
            sage: f._derivative()
            4*x^3*y^3 + O(y^4)
            sage: f._derivative(y)
            4*x^3*y^3 + O(y^4)
            sage: f._derivative(x)
            3*x^2*y^4 + O(y^5)

        TESTS::

            sage: R.<t> = PowerSeriesRing(QQ, sparse=True)
            sage: x = var('x')                                                          # needs sage.symbolic
            sage: t.derivative(x)
            Traceback (most recent call last):
            ...
            ValueError: cannot differentiate with respect to x
        """
        if var is not None and var != self._parent.gen():
            try:
                # call _derivative() recursively on coefficients
                return PowerSeries_poly(self._parent, self.__f._derivative(var),
                                    self.prec(), check=False)
            except AttributeError:
                raise ValueError('cannot differentiate with respect to {}'.format(var))

        # compute formal derivative with respect to generator
        return PowerSeries_poly(self._parent, self.__f._derivative(),
                                self.prec()-1, check=False)

    def integral(self, var=None):
        """
        Return the integral of this power series.

        By default, the integration variable is the variable of the
        power series.

        Otherwise, the integration variable is the optional parameter ``var``.

        .. NOTE::

            The integral is always chosen so the constant term is 0.

        EXAMPLES::

            sage: k.<w> = QQ[[]]
            sage: (1+17*w+15*w^3+O(w^5)).integral()
            w + 17/2*w^2 + 15/4*w^4 + O(w^6)
            sage: (w^3 + 4*w^4 + O(w^7)).integral()
            1/4*w^4 + 4/5*w^5 + O(w^8)
            sage: (3*w^2).integral()
            w^3

        TESTS::

            sage: t = PowerSeriesRing(QQ,'t').gen()
            sage: f = t + 5*t^2 + 21*t^3
            sage: g = f.integral(); g
            1/2*t^2 + 5/3*t^3 + 21/4*t^4
            sage: g.parent()
            Power Series Ring in t over Rational Field

            sage: R.<x> = QQ[]
            sage: t = PowerSeriesRing(R,'t').gen()
            sage: f = x*t +5*t^2
            sage: f.integral()
            1/2*x*t^2 + 5/3*t^3
            sage: f.integral(x)
            1/2*x^2*t + 5*x*t^2
        """
        return PowerSeries_poly(self._parent, self.__f.integral(var),
                                self.prec()+1, check=False)

    def reverse(self, precision=None):
        """
        Return the reverse of `f`, i.e., the series `g` such that `g(f(x)) = x`.

        Given an optional argument ``precision``, return the reverse with given
        precision (note that the reverse can have precision at most
        ``f.prec()``).  If `f` has infinite precision, and the argument
        ``precision`` is not given, then the precision of the reverse defaults
        to the default precision of ``f.parent()``.

        Note that this is only possible if the valuation of ``self`` is exactly
        1.

        ALGORITHM:

        We first attempt to pass the computation to pari; if this fails, we
        use Lagrange inversion.  Using ``sage: set_verbose(1)`` will print
        a message if passing to pari fails.

        If the base ring has positive characteristic, then we attempt to
        lift to a characteristic zero ring and perform the reverse there.
        If this fails, an error is raised.

        EXAMPLES::

            sage: R.<x> = PowerSeriesRing(QQ)
            sage: f = 2*x + 3*x^2 - x^4 + O(x^5)
            sage: g = f.reverse()
            sage: g
            1/2*x - 3/8*x^2 + 9/16*x^3 - 131/128*x^4 + O(x^5)
            sage: f(g)
            x + O(x^5)
            sage: g(f)
            x + O(x^5)

            sage: A.<t> = PowerSeriesRing(ZZ)
            sage: a = t - t^2 - 2*t^4 + t^5 + O(t^6)
            sage: b = a.reverse(); b
            t + t^2 + 2*t^3 + 7*t^4 + 25*t^5 + O(t^6)
            sage: a(b)
            t + O(t^6)
            sage: b(a)
            t + O(t^6)

            sage: B.<b,c> = PolynomialRing(ZZ)
            sage: A.<t> = PowerSeriesRing(B)
            sage: f = t + b*t^2 + c*t^3 + O(t^4)
            sage: g = f.reverse(); g
            t - b*t^2 + (2*b^2 - c)*t^3 + O(t^4)
            sage: f(g)
            t + O(t^4)
            sage: g(f)
            t + O(t^4)

            sage: A.<t> = PowerSeriesRing(ZZ)
            sage: B.<s> = A[[]]
            sage: f = (1 - 3*t + 4*t^3 + O(t^4))*s + (2 + t + t^2 + O(t^3))*s^2 + O(s^3)
            sage: from sage.misc.verbose import set_verbose
            sage: set_verbose(1)
            sage: g = f.reverse(); g
            verbose 1 (<module>) passing to pari failed; trying Lagrange inversion
            (1 + 3*t + 9*t^2 + 23*t^3 + O(t^4))*s + (-2 - 19*t - 118*t^2 + O(t^3))*s^2 + O(s^3)
            sage: set_verbose(0)
            sage: f(g) == g(f) == s
            True

        If the leading coefficient is not a unit, we pass to its fraction
        field if possible::

            sage: A.<t> = PowerSeriesRing(ZZ)
            sage: a = 2*t - 4*t^2 + t^4 - t^5 + O(t^6)
            sage: a.reverse()
            1/2*t + 1/2*t^2 + t^3 + 79/32*t^4 + 437/64*t^5 + O(t^6)

            sage: B.<b> = PolynomialRing(ZZ)
            sage: A.<t> = PowerSeriesRing(B)
            sage: f = 2*b*t + b*t^2 + 3*b^2*t^3 + O(t^4)
            sage: g = f.reverse(); g
            1/(2*b)*t - 1/(8*b^2)*t^2 + ((-3*b + 1)/(16*b^3))*t^3 + O(t^4)
            sage: f(g)
            t + O(t^4)
            sage: g(f)
            t + O(t^4)

        We can handle some base rings of positive characteristic::

            sage: A8.<t> = PowerSeriesRing(Zmod(8))
            sage: a = t - 15*t^2 - 2*t^4 + t^5 + O(t^6)
            sage: b = a.reverse(); b
            t + 7*t^2 + 2*t^3 + 5*t^4 + t^5 + O(t^6)
            sage: a(b)
            t + O(t^6)
            sage: b(a)
            t + O(t^6)

        The optional argument ``precision`` sets the precision of the output::

            sage: R.<x> = PowerSeriesRing(QQ)
            sage: f = 2*x + 3*x^2 - 7*x^3 + x^4 + O(x^5)
            sage: g = f.reverse(precision=3); g
            1/2*x - 3/8*x^2 + O(x^3)
            sage: f(g)
            x + O(x^3)
            sage: g(f)
            x + O(x^3)

        If the input series has infinite precision, the precision of the
        output is automatically set to the default precision of the parent
        ring::

            sage: R.<x> = PowerSeriesRing(QQ, default_prec=20)
            sage: (x - x^2).reverse() # get some Catalan numbers
            x + x^2 + 2*x^3 + 5*x^4 + 14*x^5 + 42*x^6 + 132*x^7 + 429*x^8 + 1430*x^9
             + 4862*x^10 + 16796*x^11 + 58786*x^12 + 208012*x^13 + 742900*x^14
             + 2674440*x^15 + 9694845*x^16 + 35357670*x^17 + 129644790*x^18
             + 477638700*x^19 + O(x^20)
            sage: (x - x^2).reverse(precision=3)
            x + x^2 + O(x^3)

        TESTS::

            sage: R.<x> = PowerSeriesRing(QQ)
            sage: f = 1 + 2*x + 3*x^2 - x^4 + O(x^5)
            sage: f.reverse()
            Traceback (most recent call last):
            ...
            ValueError: Series must have valuation one for reversion.

            sage: Series = PowerSeriesRing(SR, 'x')                                     # needs sage.symbolic
            sage: ser = Series([0, pi]); ser                                            # needs sage.symbolic
            pi*x
            sage: ser.reverse()                                                         # needs sage.symbolic
            1/pi*x + O(x^20)
        """
        if self.valuation() != 1:
            raise ValueError("Series must have valuation one for reversion.")

        f = self

        if f.prec() is infinity and precision is None:
            precision = f.parent().default_prec()
        if precision:
            f = f.add_bigoh(precision)

        out_prec = f.prec()

        if not f[1].is_unit():
            # if leading coefficient is not a unit, attempt passing
            # to fraction field
            try:
                f = f.change_ring(f.base_ring().fraction_field())
            except TypeError:
                raise TypeError("Leading coefficient must be a unit, or base ring must have a fraction field.")

        # set output parent after possibly passing to fraction field,
        # but before possibly lifting to characteristic zero
        out_parent = f.parent()

        # first, try reversion with pari; this is faster than Lagrange inversion
        try:
            f2 = f.__pari__()
            g = f2.serreverse()
            return PowerSeries_poly(f.parent(), g.Vec(-out_prec), out_prec)
        except (TypeError, ValueError, AttributeError, PariError, ImportError):
            # if pari fails, continue with Lagrange inversion
            from sage.misc.verbose import verbose
            verbose("passing to pari failed; trying Lagrange inversion")

        if f.parent().characteristic():
            # over a ring of positive characteristic, attempt lifting to
            # characteristic zero ring
            verbose("parent ring has positive characteristic; attempting lift to characteristic zero")
            base_lift = f.base_ring().lift().codomain()
            verbose("characteristic zero base is "+str(base_lift))
            f_lift = f.change_ring(base_lift)
            verbose("f_lift is "+str(f_lift))
            rev_lift = f_lift.reverse()
            return rev_lift.change_ring(f.base_ring())

        t = f.parent().gen()
        R = f.parent().base_ring()

        h = t/f
        k = 1
        g = 0
        for i in range(1, out_prec):
            k *= h
            g += R(k.padded_list(i)[i - 1]/i)*t**i
        g = g.add_bigoh(out_prec)
        return PowerSeries_poly(out_parent, g, out_prec, check=False)

    def pade(self, m, n):
        r"""
        Return the Padé approximant of ``self`` of index `(m, n)`.

        The Padé approximant of index `(m, n)` of a formal power
        series `f` is the quotient `Q/P` of two polynomials `Q` and `P`
        such that `\deg(Q)\leq m`, `\deg(P)\leq n` and

        .. MATH::

            f(z) - Q(z)/P(z) = O(z^{m+n+1}).

        The formal power series `f` must be known up to order `n + m`.

        See :wikipedia:`Padé\_approximant`

        INPUT:

        - ``m``, ``n`` -- integers, describing the degrees of the polynomials

        OUTPUT: a ratio of two polynomials

        ALGORITHM:

        This method uses the formula as a quotient of two determinants.

        .. SEEALSO::

            * :mod:`sage.matrix.berlekamp_massey`,
            * :meth:`sage.rings.polynomial.polynomial_zmod_flint.Polynomial_zmod_flint.rational_reconstruction`

        EXAMPLES::

            sage: z = PowerSeriesRing(QQ, 'z').gen()
            sage: exp(z).pade(4, 0)
            1/24*z^4 + 1/6*z^3 + 1/2*z^2 + z + 1
            sage: exp(z).pade(1, 1)
            (-z - 2)/(z - 2)
            sage: exp(z).pade(3, 3)
            (-z^3 - 12*z^2 - 60*z - 120)/(z^3 - 12*z^2 + 60*z - 120)
            sage: log(1-z).pade(4, 4)
            (25/6*z^4 - 130/3*z^3 + 105*z^2 - 70*z)/(z^4 - 20*z^3 + 90*z^2
            - 140*z + 70)
            sage: sqrt(1+z).pade(3, 2)
            (1/6*z^3 + 3*z^2 + 8*z + 16/3)/(z^2 + 16/3*z + 16/3)
            sage: exp(2*z).pade(3, 3)
            (-z^3 - 6*z^2 - 15*z - 15)/(z^3 - 6*z^2 + 15*z - 15)

        TESTS:

        With real coefficients::

            sage: # needs sage.rings.real_mpfr
            sage: R.<z> = RR[[]]
            sage: f = exp(2*z)
            sage: f.pade(3, 3) # abs tol 1e-10
            (-z^3 - 6.0*z^2 - 15.0*z - 15.0)/(z^3 - 6.0*z^2 + 15.0*z - 15.0)

        When precision is too low::

            sage: # needs sage.rings.real_mpfr
            sage: f = z + O(z**6)
            sage: f.pade(4, 4)
            Traceback (most recent call last):
            ...
            ValueError: the precision of the series is not large enough

        Check that :issue:`21212` is fixed::

            sage: QQx.<x> = QQ[[]]
            sage: (1 + x + O(x^100)).pade(2,2)
            x + 1

        Check for correct precision::

            sage: QQx.<x> = QQ[[]]
            sage: (1 + x + O(x^2)).pade(0,1)
            -1/(x - 1)
        """
        if self.precision_absolute() < n + m + 1:
            raise ValueError("the precision of the series is not large enough")
        polyring = self.parent()._poly_ring()
        z = polyring.gen()
        c = self.polynomial()
        u, v = c.rational_reconstruction(z**(n + m + 1), m, n)
        return u / v

    def _symbolic_(self, ring):
        """
        Conversion to symbolic series.

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: R.<x> = PowerSeriesRing(QQ)
            sage: s = R([1,2,3,4,5], prec=10); s
            1 + 2*x + 3*x^2 + 4*x^3 + 5*x^4 + O(x^10)
            sage: SR(s)
            1 + 2*x + 3*x^2 + 4*x^3 + 5*x^4 + Order(x^10)
            sage: SR(s).is_terminating_series()
            False
            sage: SR(s).variables()
            (x,)
            sage: s = R([1,2,3,4,5]); s
            1 + 2*x + 3*x^2 + 4*x^3 + 5*x^4
            sage: SR(s)
            1 + 2*x + 3*x^2 + 4*x^3 + 5*x^4
            sage: _.is_terminating_series()
            True

        TESTS:

        Check that :issue:`18094` is fixed::

            sage: R.<x> = PolynomialRing(ZZ)
            sage: SR(R(0).add_bigoh(20))                                                # needs sage.symbolic
            Order(x^20)
        """
        from sage.symbolic.ring import SR
        pex = SR(self.polynomial())
        var = SR.var(self.variable())
        return pex.series(var, self.prec())


def make_powerseries_poly_v0(parent, f, prec, is_gen):
    """
    Return the power series specified by ``f``, ``prec``, and ``is_gen``.

    This function exists for the purposes of pickling. Do not delete
    this function -- if you change the internal representation,
    instead make a new function and make sure that both kinds of
    objects correctly unpickle as the new type.

    EXAMPLES::

        sage: R.<t> = QQ[[]]
        sage: sage.rings.power_series_poly.make_powerseries_poly_v0(R, t, infinity, True)
        t
    """
    return PowerSeries_poly(parent, f, prec, 0, is_gen)


cdef class BaseRingFloorDivAction(Action):
    """
    The floor division action of the base ring on a formal power series.
    """
    cpdef _act_(self, g, x):
        r"""
        Let ``g`` act on ``x`` under ``self``.

        Regardless of whether this is a left or right action, the acting
        element comes first.

        INPUT:

        - ``g`` -- an object with parent ``self.G``
        - ``x`` -- an object with parent ``self.US()``

        .. WARNING::

            This is meant to be a fast internal function, so the
            conditions on the input are not checked!

        EXAMPLES:

        One gets the correct parent with floor division::

            sage: A = ZZ[['t']]
            sage: f = A([3*2**n for n in range(6)]).O(6)
            sage: g = f // 3; g
            1 + 2*t + 4*t^2 + 8*t^3 + 16*t^4 + 32*t^5 + O(t^6)
            sage: g.parent()
            Power Series Ring in t over Integer Ring

        whereas the parent is larger with division::

            sage: parent(f/3)
            Power Series Ring in t over Rational Field

        Floor division in case that the power series is not divisible by the divisor::

            sage: f = A([2**n for n in range(6)]).O(6)
            sage: g = f // 3; g
            t^2 + 2*t^3 + 5*t^4 + 10*t^5 + O(t^6)

        Another example::

            sage: s = polygen(QQ,'s')
            sage: A = s.parent()[['t']]
            sage: f = A([(s+2)*(s+n) for n in range(5)]).O(5)
            sage: g = f // (s + 2); g
            s + (s + 1)*t + (s + 2)*t^2 + (s + 3)*t^3 + (s + 4)*t^4 + O(t^5)
            sage: g.parent()
            Power Series Ring in t over Univariate Polynomial Ring in s
            over Rational Field

            sage: R.<t> = PowerSeriesRing(QQ)
            sage: t // 2
            1/2*t
        """
        cdef PowerSeries_poly elt = <PowerSeries_poly> x
        prec = x.prec()
        P = self.US()
        g = P.base_ring()(g)
        return type(x)(P, elt.__f // g, prec=prec, check=False)
